import matplotlib.pyplot as plt

from tf_pwa.particle import BaseDecay, BaseParticle, DecayChain
from tf_pwa.vis import *

ex_result = """
digraph {
    rankdir=LR;
    node [shape=point];
    edge [arrowhead=none, labelfloat=true];
    "A" [shape=none];
    "B" [shape=none];
    "C" [shape=none];
    "D" [shape=none];
    "E" [shape=none];
    "F" [shape=none];
    { rank=same "A" };
    { rank=same "B","D","E","F","C" };
    "A" -> "A->chain0_node_0+D";
    "A->chain0_node_0+D" -> "chain0_node_0->chain0_node_2+chain0_node_3";
    "A->chain0_node_0+D" -> "D";
    "chain0_node_0->chain0_node_2+chain0_node_3" -> "chain0_node_2->B+F";
    "chain0_node_0->chain0_node_2+chain0_node_3" -> "chain0_node_3->C+E";
    "chain0_node_2->B+F" -> "B";
    "chain0_node_2->B+F" -> "F";
    "chain0_node_3->C+E" -> "C";
    "chain0_node_3->C+E" -> "E";
}
"""


def test_dot():

    a = BaseParticle("A")
    c = BaseParticle("C")
    b = BaseParticle("B")
    d = BaseParticle("D")
    f = BaseParticle("E")
    e = BaseParticle("F")
    r = BaseParticle("R")
    BaseDecay(r, [b, d])
    BaseDecay(a, [r, c])
    g = DotGenerator(a)
    chains_0 = a.chain_decay()[0]
    source_0 = g.get_dot_source()[0]
    assert source_0 == DotGenerator.dot_chain(chains_0, True)
    assert source_0 != DotGenerator.dot_chain(chains_0, False)
    chains = DecayChain.from_particles(a, [b, c, d, e, f])
    assert len(chains) == 105  # (2*5-3)!!


def remove_same(decs):
    ret = [decs[0]]
    for i in decs[1:]:
        for j in ret:
            if i.topology_same(j):
                break
        else:
            ret.append(i)
    return ret


def test_plot():
    final = [BaseParticle(i) for i in ["C", "D", "E", "B"]]
    decs = DecayChain.from_particles("A", final)
    decs = remove_same(decs)
    plt.figure(figsize=(15, 9))
    for i in range(len(decs)):
        ax = plt.subplot(3, 5, i + 1)
        plot_decay_struct(decs[i], ax)
    plt.subplots_adjust(top=1, bottom=0, right=1, left=0, hspace=0, wspace=0)
    plt.margins(0, 0)
    plt.savefig("topo_4.png", dpi=300, pad_inches=0)



from tf_pwa.experimental import build_amp

from .test_full import gen_toy, toy_config


def test_build_amp(toy_config):
    amp = toy_config.get_amplitude()
    dg = toy_config.get_decay()
    data = toy_config.get_data("data")[0]

    amp2s = build_amp.cached_amp2s(dg, data)
    print(amp2s())

    _amp = build_amp.cached_amp(
        dg, data, matrix_method=build_amp.build_amp_matrix
    )
    print(_amp.python_function())


def test_build_amp2s(toy_config):
    # assert False

    amp = toy_config.get_amplitude()
    dg = toy_config.get_decay()
    data = toy_config.get_data("data")[0]
    amp2s = build_amp.build_amp2s(dg)
    _, cached_data = build_amp.build_angle_amp_matrix(dg, data)
    amp2s.python_function(data, cached_data)


def test_as_dict(toy_config):
    amp = toy_config.get_amplitude()
    dg = toy_config.get_decay()
    data = toy_config.get_data("data")[0]
    idx, cached_data = build_amp.build_angle_amp_matrix(dg, data)
    dic = build_amp.amp_matrix_as_dict(idx, cached_data)
    assert len(dic) == len(cached_data)



#!/usr/bin/env python

"""
Script for generating data for resolution of gauss function.
"""

import numpy as np
import tensorflow as tf

from tf_pwa.amp import get_particle, get_relative_p2
from tf_pwa.angle import LorentzVector as lv
from tf_pwa.angle import Vector3 as v3
from tf_pwa.cal_angle import cal_helicity_angle
from tf_pwa.config_loader import ConfigLoader
from tf_pwa.data import data_shape


def gauss_sample(data, decay_chain, r_name, sigma, sample_N, dat_order):
    sigma_delta = 5

    def gauss(delta_x):
        return tf.exp(-(delta_x ** 2) / (2 * sigma ** 2))

    angle = cal_helicity_angle(
        data["particle"], decay_chain.standard_topology()
    )
    decay_chain.standard_topology()
    tp_map = decay_chain.topology_map()

    r_particle = tp_map[get_particle(r_name)]

    for i in decay_chain.standard_topology():
        if i.core == r_particle:
            m_min = sum(data["particle"][j]["m"] for j in i.outs)
        print(i.outs, r_particle)
        if any(r_particle == j for j in i.outs):
            m_max = (
                data["particle"][i.core]["m"]
                - sum(data["particle"][j]["m"] for j in i.outs)
                + data["particle"][r_particle]["m"]
            )

    print("min, max: ", m_min, m_max)
    mass = {}
    weights = []
    for i in data["particle"]:
        mi = data["particle"][i]["m"]
        if i == r_particle:
            delta_min = tf.where(
                m_min - mi > -sigma_delta * sigma,
                m_min - mi,
                -sigma_delta * sigma,
            )
            delta_max = tf.where(
                m_max - mi > sigma_delta * sigma,
                sigma_delta * sigma,
                m_max - mi,
            )
            delta_m = (delta_max - delta_min) / (sample_N + 1)
            print("delta_min:", delta_min)
            min_m = mi + delta_min + delta_m / 2
            mi_s = []
            for j in range(sample_N):
                mi_s_i = min_m + delta_m * j
                mi_s.append(mi_s_i)
                weights.append(gauss(mi_s_i - mi))
            mass[i] = tf.stack(mi_s)
        else:
            mass[i] = mi[None, :]

    # print(mass[r_particle], np.mean(mass[r_particle]))

    weights = tf.stack(weights)
    weights = weights / tf.reduce_sum(weights, axis=0)
    data_weights = data.get("weight", tf.ones_like(weights))

    total_weights = weights * data_weights

    print({k: v.shape for k, v in mass.items()})

    mask = True
    p4_all = {}
    for i in decay_chain:
        phi = angle[tp_map[i]][tp_map[i.outs[0]]]["ang"]["alpha"]
        theta = angle[tp_map[i]][tp_map[i.outs[0]]]["ang"]["beta"]

        m0 = mass[tp_map[i.core]]
        m1 = mass[tp_map[i.outs[0]]]
        m2 = mass[tp_map[i.outs[1]]]

        p_square = get_relative_p2(m0, m1, m2)

        print(m0.shape, m1.shape, m2.shape, p_square.shape)

        p = tf.sqrt(tf.where(p_square > 0, p_square, 0))
        pz = p * tf.cos(theta)
        px = p * tf.sin(theta) * tf.cos(phi)
        py = p * tf.sin(theta) * tf.sin(phi)
        E1 = tf.sqrt(m1 * m1 + p * p)
        E2 = tf.sqrt(m2 * m2 + p * p)
        p1 = tf.stack([E1, px, py, pz], axis=-1)
        p2 = tf.stack([E2, -px, -py, -pz], axis=-1)
        p4_all[i.outs[0]] = p1
        p4_all[i.outs[1]] = p2

    print("p shape", {k: v.shape for k, v in p4_all.items()})

    core_boost = {}
    for i in decay_chain:
        if i.core != decay_chain.top:
            core_boost[i.outs[0]] = i.core
            core_boost[i.outs[1]] = i.core
    ret = {}
    for i in decay_chain.outs:
        tmp = i
        ret[i] = p4_all[i]
        while tmp in core_boost:
            tmp = core_boost[tmp]
            # print(i, tmp)
            print(tmp)
            ret[i] = lv.rest_vector(lv.neg(p4_all[tmp]), ret[i])

    ret2 = {}
    mask = tf.expand_dims(mask, -1)
    for i in ret:
        ret2[i] = tf.where(mask, ret[i], data["particle"][tp_map[i]]["p"])

    print("ret2:", {k: v.shape for k, v in ret2.items()})
    # print({i: data["particle"][tp_map[i]]["p"] for i in decay_chain.outs})

    pi = np.stack([ret2[i] for i in dat_order], axis=-2)
    pi = np.transpose(pi, (1, 0, 2, 3))
    total_weights = np.transpose(total_weights.numpy(), (1, 0))
    print(pi.shape)
    return pi, total_weights


def main():
    sigma = 0.005
    sigma_delta = 5
    r_name = "R_BC"

    config = ConfigLoader("config.yml")

    sample_N = config.resolution_size

    decays = config.get_decay(False)
    decay_chain = decays.get_decay_chain(r_name)
    data = config.get_data("data_origin")[0]
    pi, total_weights = gauss_sample(
        data, decay_chain, "R_BC", sigma, sample_N, config.get_dat_order()
    )
    np.savetxt("data/data.dat", pi.reshape((-1, 4)))
    np.savetxt("data/data_weight.dat", np.reshape(total_weights, (-1,)))

    data = config.get_data("phsp_plot")[0]
    pi, total_weights = gauss_sample(
        data, decay_chain, "R_BC", sigma, sample_N, config.get_dat_order()
    )
    np.save("data/phsp_re.npy", pi.reshape((-1, 4)))
    np.savetxt("data/phsp_re_weight.dat", np.reshape(total_weights, (-1,)))


if __name__ == "__main__":
    main()



"""
module for parsing decay card :code:`*.dec` file
"""
import re
import warnings

# from pysnooper import snoop


def load_dec(s):
    """
    load :code:`*.dec` string
    """
    lines = split_lines(s)
    return process_decay_card(lines)


def load_dec_file(f):
    """
    load :code:`*.dec` file
    """
    lines = f.readlines()
    return process_decay_card(lines)


def split_lines(s):
    """split each lines"""
    return re.split("\\s*\n\\s*", s)


def remove_comment(words):
    """remove comment string which starts with '#'."""
    ret = []
    for i in words:
        if len(i) <= 0:
            continue
        if i.startswith("#"):
            break
        if i.find("#") != -1:
            s = i.split("#")[0]
            ret.append(s)
            break
        ret.append(i)
    return ret


def get_words(lines):
    """get all words in a lines"""
    while True:
        line = next(lines)
        if len(line) <= 0:
            continue
        words_o = re.split(r"\s+", line)
        words = remove_comment(words_o)
        if len(words) >= 1:
            return words


commands = {}


def regist_command(name=None):
    """regist command function for command call"""

    def g(f):
        name1 = name
        if name1 is None:
            if hasattr(f, "__name__"):
                name1 = f.__name__
            elif hasattr(f, "name"):
                name1 = f.name
            else:
                raise Exception("name required")
        if name1 in commands:
            warnings.warn("override dec commands {}".format(name1))
        commands[name1] = f
        return f

    return g


def do_command(cmd, params, lines):
    """do command in commands"""
    if cmd in commands:
        fun = commands[cmd]
        return fun(params, lines)
    else:
        return cmd, params


def process_decay_card(lines):
    """process all the files as a generator"""
    lines = iter(lines)
    while True:
        words = get_words(lines)
        cmd = words[0]
        params = words[1:]
        if cmd == "End":
            break
        yield do_command(cmd, params, lines)


@regist_command(name="Decay")
def get_decay(words, lines):
    """parser decay command"""
    if len(words) <= 0:
        raise Exception("Decay need particles")
    core = words[0]
    ret = []
    while True:
        words = get_words(lines)
        if words[0] == "Enddecay":
            break
        if words[-1].endswith(";"):
            words[-1] = words[-1][:-1]
        s = sigle_decay(words)
        ret.append(s)
    return ("Decay", {"name": core, "final": ret})


models = {"HELCOV": None, "PHSP": None}


def sigle_decay(s):
    """do each decay line"""
    total = float(s[0])
    s = s[1:]
    outs = []
    model = "HELCOV"
    params = []
    is_particle = True
    for i in s:
        if i in models:
            model = i
            is_particle = False
        else:
            if is_particle:
                outs.append(i)
            else:
                params.append(i)
    return {"outs": outs, "model": model, "total": total, "params": params}


@regist_command(name="Particle")
def get_particles(words, _lines):
    """parser particles command"""
    if len(words) <= 0:
        raise Exception("Decay need particles")
    name = words[0]
    return ("Particle", {"name": name, "params": words[1:]})



import itertools
import os

import matplotlib.pyplot as plt
import numpy as np

from tf_pwa.adaptive_bins import AdaptiveBound
from tf_pwa.adaptive_bins import cal_chi2 as cal_chi2_o
from tf_pwa.data import (
    data_index,
    data_merge,
    data_shape,
    data_split,
    data_to_numpy,
    load_data,
    save_data,
)
from tf_pwa.histogram import Hist1D, interp_hist
from tf_pwa.root_io import has_uproot, save_dict_to_root

from .config_loader import ConfigLoader


def _get_cfit_bg(self, data, phsp):
    model = self._get_model()
    bg_function = [i.bg for i in model]
    w_bkg = [i.w_bkg for i in model]
    phsp_weight = []
    for data_i, phsp_i, w, bg_f in zip(data, phsp, w_bkg, bg_function):
        ndata = np.sum(data_i.get_weight())
        nbg = ndata * w
        w_bg = bg_f(phsp_i) * phsp_i.get_weight()
        phsp_weight.append(-w_bg / np.sum(w_bg) * nbg)
    return [
        type(phsp_i)({**phsp_i, "weight": w})
        for phsp_i, w in zip(phsp, phsp_weight)
    ]


def _get_cfit_eff_phsp(self, phsp):
    model = self._get_model()
    eff_function = [i.eff for i in model]
    phsp_weight = []
    for phsp_i, eff_f in zip(phsp, eff_function):
        w_eff = eff_f(phsp_i) * phsp_i.get_weight()
        phsp_weight.append(w_eff)
    return [
        type(phsp_i)({**phsp_i, "weight": w})
        for phsp_i, w in zip(phsp, phsp_weight)
    ]


@ConfigLoader.register_function()
def plot_partial_wave(
    self,
    params=None,
    data=None,
    phsp=None,
    bg=None,
    prefix="figure/",
    res=None,
    save_root=False,
    **kwargs
):
    if params is None:
        params = {}
    nll = None
    if hasattr(params, "min_nll"):
        nll = float(getattr(params, "min_nll"))
    if hasattr(params, "params"):
        params = getattr(params, "params")
    # print(nll, params)
    path = os.path.dirname(prefix)
    os.makedirs(path, exist_ok=True)

    if data is None:
        data = self.get_data("data")
        bg = self.get_data("bg")
        phsp = self.get_phsp_plot()
    if bg is None:
        if self.config["data"].get("model", "auto") == "cfit":
            bg = _get_cfit_bg(self, data, phsp)
        else:
            bg = [bg] * len(data)
    if self.config["data"].get("model", "auto") == "cfit":
        phsp = _get_cfit_eff_phsp(self, phsp)
    amp = self.get_amplitude()
    self._Ngroup = len(data)
    ws_bkg = [
        None if bg_i is None else bg_i.get("weight", None) for bg_i in bg
    ]
    # ws_bkg, ws_inmc = self._get_bg_weight(data, bg)
    chain_property = []
    if res is None:
        for i in range(len(self.full_decay.chains)):
            name_i, curve_style = self.get_chain_property(i, False)
            label, curve_style = self.get_chain_property(i, True)
            chain_property.append([i, name_i, label, curve_style])
    else:
        for i, name in enumerate(res):
            if not isinstance(name, list):
                name = [name]
            if len(name) == 1:
                display = str(name[0])
            else:
                display = "{ " + ",\n  ".join([str(i) for i in name]) + " }"
            name_i = "_".join([str(i) for i in name])
            chain_property.append([i, name_i, display, None])
    plot_var_dic = {}
    for conf in self.plot_params.get_params():
        name = conf.get("name")
        display = conf.get("display", name)
        upper_ylim = conf.get("upper_ylim", None)
        idx = conf.get("idx")
        trans = conf.get("trans", lambda x: x)
        has_legend = conf.get("legend", False)
        xrange = conf.get("range", None)
        bins = conf.get("bins", None)
        units = conf.get("units", "")
        yscale = conf.get("yscale", "linear")
        plot_var_dic[name] = {
            "display": display,
            "upper_ylim": upper_ylim,
            "legend": has_legend,
            "idx": idx,
            "trans": trans,
            "range": xrange,
            "bins": bins,
            "units": units,
            "yscale": yscale,
        }
    if self._Ngroup == 1:
        data_dict, phsp_dict, bg_dict = self._cal_partial_wave(
            amp,
            params,
            data[0],
            phsp[0],
            bg[0],
            ws_bkg[0],
            prefix,
            plot_var_dic,
            chain_property,
            save_root=save_root,
            res=res,
            **kwargs,
        )
        self._plot_partial_wave(
            data_dict,
            phsp_dict,
            bg_dict,
            prefix,
            plot_var_dic,
            chain_property,
            nll=nll,
            **kwargs,
        )
    else:
        combine_plot = self.config["plot"].get("combine_plot", True)
        if not combine_plot:
            for dt, mc, sb, w_bkg, i in zip(
                data, phsp, bg, ws_bkg, range(self._Ngroup)
            ):
                data_dict, phsp_dict, bg_dict = self._cal_partial_wave(
                    amp,
                    params,
                    dt,
                    mc,
                    sb,
                    w_bkg,
                    prefix + "d{}_".format(i),
                    plot_var_dic,
                    chain_property,
                    save_root=save_root,
                    **kwargs,
                )
                self._plot_partial_wave(
                    data_dict,
                    phsp_dict,
                    bg_dict,
                    prefix + "d{}_".format(i),
                    plot_var_dic,
                    chain_property,
                    nll=nll,
                    **kwargs,
                )
        else:

            for dt, mc, sb, w_bkg, i in zip(
                data, phsp, bg, ws_bkg, range(self._Ngroup)
            ):
                data_dict, phsp_dict, bg_dict = self._cal_partial_wave(
                    amp,
                    params,
                    dt,
                    mc,
                    sb,
                    w_bkg,
                    prefix + "d{}_".format(i),
                    plot_var_dic,
                    chain_property,
                    save_root=save_root,
                    res=res,
                    **kwargs,
                )
                # self._plot_partial_wave(data_dict, phsp_dict, bg_dict, path+'d{}_'.format(i), plot_var_dic, chain_property, **kwargs)
                if i == 0:
                    datas_dict = {}
                    for ct in data_dict:
                        datas_dict[ct] = [data_dict[ct]]
                    phsps_dict = {}
                    for ct in phsp_dict:
                        phsps_dict[ct] = [phsp_dict[ct]]
                    bgs_dict = {}
                    for ct in bg_dict:
                        bgs_dict[ct] = [bg_dict[ct]]
                else:
                    for ct in data_dict:
                        datas_dict[ct].append(data_dict[ct])
                    for ct in phsp_dict:
                        phsps_dict[ct].append(phsp_dict[ct])
                    for ct in bg_dict:
                        bgs_dict[ct].append(bg_dict[ct])
            for ct in datas_dict:
                datas_dict[ct] = np.concatenate(datas_dict[ct])
            for ct in phsps_dict:
                phsps_dict[ct] = np.concatenate(phsps_dict[ct])
            for ct in bgs_dict:
                bgs_dict[ct] = np.concatenate(bgs_dict[ct])
            self._plot_partial_wave(
                datas_dict,
                phsps_dict,
                bgs_dict,
                prefix + "com_",
                plot_var_dic,
                chain_property,
                nll=nll,
                **kwargs,
            )
            if has_uproot and save_root:
                if bg[0] is None:
                    save_dict_to_root(
                        [datas_dict, phsps_dict],
                        file_name=prefix + "variables_com.root",
                        tree_name=["data", "fitted"],
                    )
                else:
                    save_dict_to_root(
                        [datas_dict, phsps_dict, bgs_dict],
                        file_name=prefix + "variables_com.root",
                        tree_name=["data", "fitted", "sideband"],
                    )
                print("Save root file " + prefix + "com_variables.root")


@ConfigLoader.register_function()
def _cal_partial_wave(
    self,
    amp,
    params,
    data,
    phsp,
    bg,
    w_bkg,
    prefix,
    plot_var_dic,
    chain_property,
    save_root=False,
    bin_scale=3,
    res=None,
    batch=65000,
    **kwargs
):
    data_dict = {}
    phsp_dict = {}
    bg_dict = {}
    with amp.temp_params(params):
        weights_i = [amp(i) for i in data_split(phsp, batch)]
        weight_phsp = data_merge(*weights_i)  # amp(phsp)
        total_weight = (
            weight_phsp * phsp.get("weight", 1.0) * phsp.get("eff_value", 1.0)
        )
        data_weight = data.get("weight", None)
        if data_weight is None:
            n_data = data_shape(data)
        else:
            n_data = np.sum(data_weight)
        if bg is None:
            norm_frac = n_data / np.sum(total_weight)
        else:
            if isinstance(w_bkg, float):
                norm_frac = (n_data - w_bkg * data_shape(bg)) / np.sum(
                    total_weight
                )
            else:
                norm_frac = (n_data + np.sum(w_bkg)) / np.sum(total_weight)
        if res is None:
            weights = amp.partial_weight(phsp)
        else:
            weights = []
            used_res = amp.used_res
            for i in res:
                if not isinstance(i, list):
                    i = [i]
                amp.set_used_res(i)
                weights.append(amp(phsp))
            # print(weights, amp.decay_group.chains_idx)
            amp.set_used_res(used_res)

        data_weights = data.get("weight", np.ones((data_shape(data),)))
        data_dict["data_weights"] = data_weights
        phsp_weights = total_weight * norm_frac
        phsp_dict["MC_total_fit"] = phsp_weights  # MC total weight
        if bg is not None:
            if isinstance(w_bkg, float):
                bg_weight = [w_bkg] * data_shape(bg)
            else:
                bg_weight = -w_bkg
            bg_dict["sideband_weights"] = bg_weight  # sideband weight
        for i, name_i, label, _ in chain_property:
            weight_i = (
                weights[i]
                * norm_frac
                * bin_scale
                * phsp.get("weight", 1.0)
                * phsp.get("eff_value", 1.0)
            )
            phsp_dict[
                "MC_{0}_{1}_fit".format(i, name_i)
            ] = weight_i  # MC partial weight
        for name in plot_var_dic:
            idx = plot_var_dic[name]["idx"]
            trans = lambda x: np.reshape(plot_var_dic[name]["trans"](x), (-1,))

            data_i = trans(data_index(data, idx))
            if idx[-1] == "m":
                tmp_idx = list(idx)
                tmp_idx[-1] = "p"
                p4 = data_index(data, tmp_idx)
                p4 = np.transpose(p4)
                data_dict[name + "_E"] = p4[0]
                data_dict[name + "_PX"] = p4[1]
                data_dict[name + "_PY"] = p4[2]
                data_dict[name + "_PZ"] = p4[3]
            data_dict[name] = data_i  # data variable

            phsp_i = trans(data_index(phsp, idx))
            phsp_dict[name + "_MC"] = phsp_i  # MC

            if bg is not None:
                bg_i = trans(data_index(bg, idx))
                bg_dict[name + "_sideband"] = bg_i  # sideband
    data_dict = data_to_numpy(data_dict)
    phsp_dict = data_to_numpy(phsp_dict)
    bg_dict = data_to_numpy(bg_dict)
    if has_uproot and save_root:
        if bg is None:
            save_dict_to_root(
                [data_dict, phsp_dict],
                file_name=prefix + "variables.root",
                tree_name=["data", "fitted"],
            )
        else:
            save_dict_to_root(
                [data_dict, phsp_dict, bg_dict],
                file_name=prefix + "variables.root",
                tree_name=["data", "fitted", "sideband"],
            )
        print("Save root file " + prefix + "variables.root")
    return data_dict, phsp_dict, bg_dict


@ConfigLoader.register_function()
def _plot_partial_wave(
    self,
    data_dict,
    phsp_dict,
    bg_dict,
    prefix,
    plot_var_dic,
    chain_property,
    plot_delta=False,
    plot_pull=False,
    save_pdf=False,
    bin_scale=3,
    single_legend=False,
    format="png",
    nll=None,
    smooth=True,
    color_first=True,
    **kwargs
):

    # cmap = plt.get_cmap("jet")
    # N = 10
    # colors = [cmap(float(i) / (N+1)) for i in range(1, N+1)]
    colors = [
        "red",
        "orange",
        "purple",
        "springgreen",
        "y",
        "green",
        "blue",
        "c",
    ]
    linestyles = ["-", "--", "-.", ":"]

    data_weights = data_dict["data_weights"]
    if bg_dict:
        bg_weight = bg_dict["sideband_weights"]
    phsp_weights = phsp_dict["MC_total_fit"]
    for name in plot_var_dic:
        data_i = data_dict[name]
        phsp_i = phsp_dict[name + "_MC"]
        if bg_dict:
            bg_i = bg_dict[name + "_sideband"]

        display = plot_var_dic[name]["display"]
        upper_ylim = plot_var_dic[name]["upper_ylim"]
        has_legend = plot_var_dic[name]["legend"]
        bins = plot_var_dic[name]["bins"]
        units = plot_var_dic[name]["units"]
        xrange = plot_var_dic[name]["range"]
        yscale = plot_var_dic[name].get("yscale", "linear")
        if xrange is None:
            xrange = [np.min(data_i) - 0.1, np.max(data_i) + 0.1]
        # data_x, data_y, data_err = hist_error(
        # data_i, bins=bins, weights=data_weights, xrange=xrange
        # )
        data_hist = Hist1D.histogram(
            data_i, weights=data_weights, range=xrange, bins=bins
        )
        fig = plt.figure()
        if plot_delta or plot_pull:
            ax = plt.subplot2grid((4, 1), (0, 0), rowspan=3)
        else:
            ax = fig.add_subplot(1, 1, 1)

        legends = []
        legends_label = []

        le = data_hist.draw_error(
            ax, fmt=".", zorder=-2, label="data", color="black"
        )

        legends.append(le)
        legends_label.append("data")

        fitted_hist = Hist1D.histogram(
            phsp_i, weights=phsp_weights, range=xrange, bins=bins
        )

        if bg_dict:
            bg_hist = Hist1D.histogram(
                bg_i, weights=bg_weight, range=xrange, bins=bins
            )
            le = bg_hist.draw_bar(
                ax, label="back ground", alpha=0.5, color="grey"
            )
            fitted_hist = fitted_hist + bg_hist
            legends.append(le)
            legends_label.append("back ground")

        le2 = fitted_hist.draw(ax, label="total fit", color="black")
        legends.append(le2[0])
        legends_label.append("total fit")
        if color_first:
            style = itertools.product(linestyles, colors)
        else:
            style = itertools.product(colors, linestyles)
        for i, name_i, label, curve_style in chain_property:
            weight_i = phsp_dict["MC_{0}_{1}_fit".format(i, name_i)]
            hist_i = Hist1D.histogram(
                phsp_i,
                weights=weight_i,
                range=xrange,
                bins=bins * bin_scale,
            )
            if smooth:
                if curve_style is None:
                    if color_first:
                        ls, color = next(style)
                    else:
                        color, ls = next(style)
                    le3 = hist_i.draw_kde(
                        ax,
                        label=label,
                        color=color,
                        linestyle=ls,
                        linewidth=1,
                    )
                else:
                    le3 = hist_i.draw_kde(
                        ax, fmt=curve_style, label=label, linewidth=1
                    )
            else:
                if curve_style is None:
                    if color_first:
                        ls, color = next(style)
                    else:
                        color, ls = next(style)
                    le3 = hist_i.draw(
                        ax,
                        label=label,
                        color=color,
                        linestyle=ls,
                        linewidth=1,
                    )
                else:
                    le3 = hist_i.draw(
                        ax,
                        color=curve_style[0],
                        linestyle=curve_style[1:],
                        label=label,
                        linewidth=1,
                    )
            legends.append(le3[0])
            legends_label.append(label)
        if yscale == "log":
            ax.set_ylim((0.1, upper_ylim))
        else:
            ax.set_ylim((0, upper_ylim))
        ax.set_xlim(xrange)
        ax.set_yscale(yscale)
        if has_legend:
            leg = ax.legend(
                legends,
                legends_label,
                frameon=False,
                labelspacing=0.1,
                borderpad=0.0,
            )
        if nll is None:
            ax.set_title(display, fontsize="xx-large")
        else:
            ax.set_title(
                "{}: -lnL= {:.5}".format(display, nll), fontsize="xx-large"
            )
        ax.set_xlabel(display + units)
        ywidth = np.mean(
            data_hist.bin_width
        )  # (max(data_x) - min(data_x)) / bins
        ax.set_ylabel("Events/{:.3f}{}".format(ywidth, units))
        if plot_delta or plot_pull:
            plt.setp(ax.get_xticklabels(), visible=False)
            ax2 = plt.subplot2grid((4, 1), (3, 0), rowspan=1)
            # y_err = fit_y - data_y
            # if plot_pull:
            # _epsilon = 1e-10
            # with np.errstate(divide="ignore", invalid="ignore"):
            # fit_err = np.sqrt(fit_y)
            # y_err = y_err / fit_err
            # y_err[fit_err < _epsilon] = 0.0
            # ax2.bar(data_x, y_err, color="k", alpha=0.7, width=ywidth)
            if plot_pull:
                (data_hist - fitted_hist).draw_pull()
                ax2.axhline(y=0, color="r", linewidth=0.5)
                ax2.axhline(
                    y=3,
                    color="r",
                    linestyle="--",
                    linewidth=0.5,
                )
                ax2.axhline(
                    y=-3,
                    color="r",
                    linestyle="--",
                    linewidth=0.5,
                )
                ax2.set_ylabel("pull")
                ax2.set_ylim((-5, 5))
            else:
                diff_hist = data_hist - fitted_hist
                diff_hist.draw_bar(color="grey")
                ax2.set_ylabel("$\\Delta$Events")
                y_err = diff_hist.count
                ax2.set_ylim((-max(abs(y_err)), max(abs(y_err))))
            ax.set_xlabel("")
            ax2.set_xlabel(display + units)
            if xrange is not None:
                ax2.set_xlim(xrange)
        # ax.set_yscale("log")
        # ax.set_ylim([0.1, 1e3])
        fig.savefig(prefix + name + "." + format, dpi=300)
        if single_legend:
            export_legend(ax, prefix + "legend.{}".format(format))
        if save_pdf:
            fig.savefig(prefix + name + ".pdf", dpi=300)
            if single_legend:
                export_legend(ax, prefix + "legend.pdf")
        print("Finish plotting " + prefix + name)
        plt.close(fig)

    self._2d_plot(
        data_dict,
        phsp_dict,
        bg_dict,
        prefix,
        plot_var_dic,
        chain_property,
        plot_delta=plot_delta,
        plot_pull=plot_pull,
        save_pdf=save_pdf,
        bin_scale=bin_scale,
        single_legend=single_legend,
        format=format,
        nll=nll,
        smooth=smooth,
        color_first=color_first,
        **kwargs,
    )


@ConfigLoader.register_function()
def _2d_plot(
    self,
    data_dict,
    phsp_dict,
    bg_dict,
    prefix,
    plot_var_dic,
    chain_property,
    plot_delta=False,
    plot_pull=False,
    save_pdf=False,
    bin_scale=3,
    single_legend=False,
    format="png",
    nll=None,
    smooth=True,
    color_first=True,
    **kwargs
):

    twodplot = self.config["plot"].get("2Dplot", {})
    for k, i in twodplot.items():
        var1, var2 = k.split("&")
        var1 = var1.rstrip()
        var2 = var2.lstrip()
        k = var1 + "_vs_" + var2
        display = i.get("display", k)
        plot_figs = i["plot_figs"]
        name1, name2 = display.split("vs")
        name1 = name1.rstrip()
        name2 = name2.lstrip()
        range1 = plot_var_dic[var1]["range"]
        data_1 = data_dict[var1]
        phsp_1 = phsp_dict[var1 + "_MC"]
        range2 = plot_var_dic[var2]["range"]
        data_2 = data_dict[var2]
        phsp_2 = phsp_dict[var2 + "_MC"]

        # data
        if "data" in plot_figs:
            plt.scatter(data_1, data_2, s=1, alpha=0.8, label="data")
            plt.xlabel(name1)
            plt.ylabel(name2)
            plt.title(display, fontsize="xx-large")
            plt.legend()
            plt.xlim(range1)
            plt.ylim(range2)
            plt.savefig(prefix + k + "_data")
            plt.clf()
            print("Finish plotting 2D data " + prefix + k)
        # sideband
        if "sideband" in plot_figs:
            if bg_dict:
                bg_1 = bg_dict[var1 + "_sideband"]
                bg_2 = bg_dict[var2 + "_sideband"]
                plt.scatter(
                    bg_1, bg_2, s=1, c="g", alpha=0.8, label="sideband"
                )
                plt.xlabel(name1)
                plt.ylabel(name2)
                plt.title(display, fontsize="xx-large")
                plt.legend()
                plt.xlim(range1)
                plt.ylim(range2)
                plt.savefig(prefix + k + "_bkg")
                plt.clf()
                print("Finish plotting 2D sideband " + prefix + k)
            else:
                print("There's no bkg input")
        # fit pdf
        if "fitted" in plot_figs:
            phsp_weights = phsp_dict["MC_total_fit"]
            plt.hist2d(
                phsp_1, phsp_2, bins=100, weights=phsp_weights, cmin=1e-12
            )
            plt.xlabel(name1)
            plt.ylabel(name2)
            plt.title(display, fontsize="xx-large")
            plt.colorbar()
            plt.xlim(range1)
            plt.ylim(range2)
            plt.savefig(prefix + k + "_fitted")
            plt.clf()
            print("Finish plotting 2D fitted " + prefix + k)


def hist_error(data, bins=50, xrange=None, weights=1.0, kind="poisson"):
    if not hasattr(weights, "__len__"):
        weights = [weights] * data.__len__()
    data_hist = np.histogram(data, bins=bins, weights=weights, range=xrange)
    # ax.hist(fd(data[idx].numpy()),range=xrange,bins=bins,histtype="step",label="data",zorder=99,color="black")
    data_y, data_x = data_hist[0:2]
    data_x = (data_x[:-1] + data_x[1:]) / 2
    if kind == "poisson":
        data_err = np.sqrt(
            np.abs(data_y)
        )  # data_err = np.maximum(np.sqrt(np.abs(data_y)),1)
    elif kind == "binomial":
        n = data.shape[0]
        p = data_y / n
        data_err = np.sqrt(p * (1 - p) * n)
    else:
        raise ValueError("unknown error kind {}".format(kind))
    return data_x, data_y, data_err


def hist_line(
    data, weights, bins, xrange=None, inter=1, kind="UnivariateSpline"
):
    """interpolate data from hostgram into a line"""
    y, x = np.histogram(data, bins=bins, range=xrange, weights=weights)
    num = data.shape[0] * inter
    return interp_hist(x, y, num=num, kind=kind)


def hist_line_step(
    data, weights, bins, xrange=None, inter=1, kind="quadratic"
):
    y, x = np.histogram(data, bins=bins, range=xrange, weights=weights)
    dx = x[1] - x[0]
    x = (x[:-1] + x[1:]) / 2
    x = np.concatenate([[x[0] - dx], x, [x[-1] + dx]])
    y = np.concatenate([[0], y, [0]])
    return x, y


def export_legend(ax, filename="legend.pdf", ncol=1):
    fig2 = plt.figure()
    ax2 = fig2.add_subplot()
    ax2.axis("off")
    legend = ax2.legend(
        *ax.get_legend_handles_labels(), frameon=False, loc="lower center"
    )
    fig = legend.figure
    fig.canvas.draw()
    bbox = legend.get_window_extent().transformed(
        fig.dpi_scale_trans.inverted()
    )
    fig.savefig(filename, dpi="figure", bbox_inches=bbox)
    plt.close(fig2)
    plt.close(fig)



import os.path
import sys

import numpy as np

# import tf_pwa
from tf_pwa.config_loader import ConfigLoader
from tf_pwa.utils import error_print

this_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, this_dir + "/..")


def main():
    """Fit the amplitude model of data, calculating the fitting parameters and the fit fractions."""
    import argparse

    parser = argparse.ArgumentParser(
        description="a simple but complete analysis process"
    )
    parser.add_argument(
        "--final_params", default="final_params.json", dest="final_params"
    )
    results = parser.parse_args()

    fit(final_params_file=results.final_params)


def fit(final_params_file):
    config = ConfigLoader(
        "config.yml"
    )  # We use ConfigLoader to read the information in the configuration file
    # config.set_params("gen_params.json") # If not set, we will use random initial parameters
    fit_result = config.fit(method="BFGS")

    errors = config.get_params_error(
        fit_result
    )  # calculate Hesse errors of the parameters
    print("\n########## fit parameters:")
    for key, value in config.get_params().items():
        print(key, error_print(value, errors.get(key, None)))

    fit_result.save_as(final_params_file)  # save fit_result to a json file
    config.plot_partial_wave(
        fit_result
    )  # Plot distributions of variables indicated in the configuration file

    fit_frac, err_frac = config.cal_fitfractions()
    print("\n########## fit fractions:")
    for i in fit_frac:
        if not isinstance(i, tuple):  # fit fraction
            name = i
        else:
            name = "{}x{}".format(*i)  # interference term
        print(name + ": " + error_print(fit_frac[i], err_frac.get(i, None)))


if __name__ == "__main__":
    main()



import contextlib

import numpy as np
import tensorflow as tf


class ParamsTrans:
    def __init__(self, vm, err_matrix):
        self.vm = vm
        self.err_matrix = err_matrix
        self.tape = None

    @contextlib.contextmanager
    def trans(self):
        with tf.GradientTape(persistent=True) as tape:
            yield self
        self.tape = tape

    def get_error(self, vals, keep=False):
        if isinstance(vals, (list, tuple)):
            ret = type(vals)([self.get_error(v, keep=True) for v in vals])
        elif isinstance(vals, dict):
            ret = {k: self.get_error(v, keep=True) for k, v in vals.items()}
        elif isinstance(vals, tf.Tensor):
            if len(vals.shape) == 0:  # scalar
                grad = self.tape.gradient(
                    vals,
                    self.vm.trainable_variables,
                    unconnected_gradients="zero",
                )
                grad = tf.stack(grad, axis=-1)
                ret = tf.sqrt(
                    tf.reduce_sum(
                        tf.linalg.matvec(self.err_matrix, grad) * grad
                    )
                )
            else:  # vector
                grad = self.tape.jacobian(
                    vals,
                    self.vm.trainable_variables,
                    unconnected_gradients="zero",
                )
                grad = tf.stack(grad, axis=-1)
                grad = tf.reshape(grad, (-1, len(self.vm.trainable_variables)))
                new_err_matrix = tf.matmul(
                    tf.matmul(grad, self.err_matrix), grad, transpose_b=True
                )
                ret = tf.sqrt(tf.linalg.tensor_diag_part(new_err_matrix))
                ret = tf.reshape(ret, vals.shape)
        else:
            raise TypeError(
                f"unsuported type {type(vals)}, use tensor instead"
            )
        if not keep:
            del self.tape
        return ret

    def get_error_matrix(self, vals, keep=False):
        if isinstance(vals, (list, tuple)):
            grad = [
                self.tape.gradient(
                    i,
                    self.vm.trainable_variables,
                    unconnected_gradients="zero",
                )
                for i in vals
            ]
        elif isinstance(vals, tf.Tensor):
            grad = self.tape.jacobian(
                vals, self.vm.trainable_variables, unconnected_gradients="zero"
            )
        if not keep:
            del self.tape
        # print(grad)
        grad = np.stack(grad).reshape((-1, len(self.vm.trainable_variables)))
        # print(grad, self.err_matrix, np.dot(grad, self.err_matrix), grad.T)
        return np.dot(np.dot(grad, self.err_matrix), grad.T)

    def __getitem__(self, key):
        return self.vm.variables[key]



"""
Basic amplitude model
"""


import numpy as np

from tf_pwa.breit_wigner import BW, BWR, BWR2, GS, Bprime, BWR_normal, Gamma
from tf_pwa.breit_wigner import barrier_factor2 as barrier_factor
from tf_pwa.dec_parser import load_dec_file
from tf_pwa.dfun import get_D_matrix_lambda
from tf_pwa.particle import _spin_int, _spin_range
from tf_pwa.tensorflow_wrapper import tf

from .core import (
    AmpBase,
    AmpDecay,
    HelicityDecay,
    Particle,
    get_relative_p,
    regist_decay,
    regist_particle,
)


@regist_particle("BWR2")
class ParticleBWR2(Particle):
    """
    .. math::
        R(m) = \\frac{1}{m_0^2 - m^2 - i m_0 \\Gamma(m)}

    """

    def get_amp(self, data, data_c, **kwargs):
        mass = self.get_mass()
        width = self.get_width()
        if width is None:
            return tf.ones_like(data["m"])
        if not self.running_width:
            ret = BW(data["m"], mass, width)
        else:
            q2 = data_c["|q|2"]
            q02 = data_c["|q0|2"]
            if self.bw_l is None:
                decay = self.decay[0]
                self.bw_l = min(decay.get_l_list())
            ret = BWR2(data["m"], mass, width, q2, q02, self.bw_l, self.d)
        return ret


@regist_particle("BWR_normal")
class ParticleBWR_normal(Particle):
    """
    .. math::
        R(m) = \\frac{\\sqrt{m_0 \\Gamma(m)}}{m_0^2 - m^2 - i m_0 \\Gamma(m)}

    """

    def get_amp(self, data, data_c, **kwargs):
        mass = self.get_mass()
        width = self.get_width()
        if width is None:
            return tf.ones_like(data["m"])
        if not self.running_width:
            ret = BW(data["m"], mass, width)
        else:
            q2 = data_c["|q|2"]
            q02 = data_c["|q0|2"]
            if self.bw_l is None:
                decay = self.decay[0]
                self.bw_l = min(decay.get_l_list())
            ret = BWR_normal(
                data["m"], mass, width, q2, q02, self.bw_l, self.d
            )
        return ret


# added by xiexh for GS model rho
@regist_particle("GS_rho")
class ParticleGS(Particle):
    r"""
    Gounaris G.J., Sakurai J.J., Phys. Rev. Lett., 21 (1968), pp. 244-247

    `c_daug2Mass`: mass for daughter particle 2 (:math:`\pi^{+}`) 0.13957039

    `c_daug3Mass`: mass for daughter particle 3 (:math:`\pi^{0}`) 0.1349768

    .. math::
      R(m) = \frac{1 + D \Gamma_0 / m_0}{(m_0^2 -m^2) + f(m) - i m_0 \Gamma(m)}

    .. math::
      f(m) = \Gamma_0 \frac{m_0 ^2 }{q_0^3} \left[q^2 [h(m)-h(m_0)] + (m_0^2 - m^2) q_0^2 \frac{d h}{d m}|_{m0} \right]

    .. math::
      h(m) = \frac{2}{\pi} \frac{q}{m} \ln \left(\frac{m+q}{2m_{\pi}} \right)

    .. math::
      \frac{d h}{d m}|_{m0} = h(m_0) [(8q_0^2)^{-1} - (2m_0^2)^{-1}] + (2\pi m_0^2)^{-1}

    .. math::
      D = \frac{f(0)}{\Gamma_0 m_0} = \frac{3}{\pi}\frac{m_\pi^2}{q_0^2} \ln \left(\frac{m_0 + 2q_0}{2 m_\pi }\right)
        + \frac{m_0}{2\pi q_0} - \frac{m_\pi^2 m_0}{\pi q_0^3}
    """

    def __init__(self, *args, **kwargs):
        self.c_daug2Mass = 0.13957039
        self.c_daug3Mass = 0.1349768
        super().__init__(*args, **kwargs)

    def get_amp(self, data, data_c, **kwargs):
        mass = self.get_mass()
        width = self.get_width()
        if width is None:
            return tf.ones_like(data["m"])
        if not self.running_width:
            ret = BW(data["m"], mass, width)
        else:
            q = data_c["|q|"]
            q0 = data_c["|q0|"]
            if self.bw_l is None:
                decay = self.decay[0]
                self.bw_l = min(decay.get_l_list())
            ret = GS(
                data["m"],
                mass,
                width,
                q,
                q0,
                self.bw_l,
                self.d,
                self.c_daug2Mass,
                self.c_daug3Mass,
            )
        return ret


# added by xiexh end


@regist_particle("BW")
class ParticleBW(Particle):
    """
    .. math::
        R(m) = \\frac{1}{m_0^2 - m^2 - i m_0 \\Gamma_0}

    """

    def get_amp(self, data, _data_c=None, **kwargs):
        mass = self.get_mass()
        width = self.get_width()
        ret = BW(data["m"], mass, width)
        return ret


@regist_particle("Kmatrix")
class ParticleKmatrix(Particle):
    def init_params(self):
        self.d = 3.0
        self.mass1 = self.add_var("mass1", fix=True)
        self.mass2 = self.add_var("mass2", fix=True)
        self.width1 = self.add_var("width1", fix=True)
        self.width2 = self.add_var("width2", fix=True)
        self.KNR = self.add_var("KNR", is_complex=True)
        self.alpha = self.add_var("alpha")
        self.beta0 = self.add_var("beta0", is_complex=True)
        self.beta1 = self.add_var("beta1", is_complex=True, fix=True)
        self.beta2 = self.add_var("beta2", is_complex=True)
        if self.bw_l is None:
            decay = self.decay[0]
            self.bw_l = min(decay.get_l_list())

    def get_amp(self, data, data_c=None, **kwargs):
        m = data["m"]
        mass1 = self.mass1()
        mass2 = self.mass2()
        width1 = self.width1()
        width2 = self.width2()
        q = data_c["|q|"]
        mdaughter1 = kwargs["all_data"]["particle"][self.decay[0].outs[0]]["m"]
        mdaughter2 = kwargs["all_data"]["particle"][self.decay[0].outs[1]]["m"]
        q1 = get_relative_p(mass1, mdaughter1, mdaughter2)
        q2 = get_relative_p(mass2, mdaughter1, mdaughter2)
        mlist = tf.stack([mass1, mass2])
        wlist = tf.stack([width1, width2])
        qlist = tf.stack([q1, q2])
        Klist = []
        for mi, wi, qi in zip(mlist, wlist, qlist):
            rw = Gamma(m, wi, q, qi, self.bw_l, mi, self.d)
            Klist.append(mi * rw / (mi ** 2 - m ** 2))
        KK = tf.reduce_sum(Klist, axis=0)
        KK += self.alpha()
        beta_term = self.get_beta(
            m=m,
            mlist=mlist,
            wlist=wlist,
            q=q,
            qlist=qlist,
            Klist=Klist,
            **kwargs,
        )
        MM = tf.complex(np.float64(1), -KK)
        MM = beta_term / MM
        return MM + self.KNR()

    def get_beta(self, m, **kwargs):
        m1, m2 = kwargs["mlist"]
        w1, w2 = kwargs["wlist"]
        q1, q2 = kwargs["qlist"]
        q = kwargs["q"]
        z = (q * self.d) ** 2
        z1 = (q1 * self.d) ** 2
        z2 = (q2 * self.d) ** 2
        Klist = kwargs["Klist"]
        beta1 = self.beta1()
        beta1 = beta1 * tf.cast(Klist[0] * m / m1 * q1 / q, beta1.dtype)
        beta1 = beta1 / tf.cast(
            (z / z1) ** self.bw_l * Bprime(self.bw_l, q, q1, self.d) ** 2,
            beta1.dtype,
        )
        beta2 = self.beta2()
        beta2 = beta2 * tf.cast(Klist[1] * m / m2 * q2 / q, beta2.dtype)
        beta2 = beta2 / tf.cast(
            (z / z2) ** self.bw_l * Bprime(self.bw_l, q, q2, self.d) ** 2,
            beta2.dtype,
        )
        beta0 = self.beta0()  # * tf.cast(2 * z / (z + 1), beta1.dtype)
        return beta0 + beta1 + beta2


@regist_particle("LASS")
class ParticleLass(Particle):
    def init_params(self):
        super(ParticleLass, self).init_params()
        self.a = self.add_var("a")
        self.r = self.add_var("r")

    def get_amp(self, data, data_c=None, **kwargs):
        r"""
        .. math::
          R(m) = \frac{m}{q cot \delta_B - i q}
            + e^{2i \delta_B}\frac{m_0 \Gamma_0 \frac{m_0}{q_0}}
                                  {(m_0^2 - m^2) - i m_0\Gamma_0 \frac{q}{m}\frac{m_0}{q_0}}

        .. math::
          cot \delta_B = \frac{1}{a q} + \frac{1}{2} r q

        .. math::
          e^{2i\delta_B} = \cos 2 \delta_B + i \sin 2\delta_B
                         = \frac{cot^2\delta_B -1 }{cot^2 \delta_B +1} + i \frac{2 cot \delta_B }{cot^2 \delta_B +1 }

        """
        m = data["m"]
        q = data_c["|q|"]
        q0 = data_c["|q0|"]
        mass = self.get_mass()
        width = self.get_width()
        a, r = tf.abs(self.a()), tf.abs(self.r())
        cot_delta_B = (1.0 / a) / q + 0.5 * r * q
        cot2_delta_B = cot_delta_B * cot_delta_B
        expi_2delta_B = tf.complex(cot2_delta_B - 1, 2 * cot_delta_B)
        expi_2delta_B /= tf.cast(cot2_delta_B + 1, expi_2delta_B.dtype)
        ret = 1.0 / tf.complex(q * cot_delta_B, -q)
        ret = tf.cast(m, ret.dtype) * ret
        ret += (
            expi_2delta_B
            * BWR(m, mass, width, q, q0, 0, 1.0)
            * tf.cast(mass * width * mass / q0, ret.dtype)
        )
        return ret


@regist_particle("one")
class ParticleOne(Particle):
    """
    .. math::
        R(m) = 1

    """

    def init_params(self):
        pass

    def get_amp(self, data, _data_c=None, **kwargs):
        mass = data["m"]
        zeros = tf.zeros_like(mass)
        ones = tf.ones_like(mass)
        return tf.complex(ones, zeros)


@regist_particle("exp")
class ParticleExp(Particle):
    """
    .. math::
        R(m) = e^{-|a| m}

    """

    def init_params(self):
        self.a = self.add_var("a")

    def get_amp(self, data, _data_c=None, **kwargs):
        mass = data["m"]
        zeros = tf.zeros_like(mass)
        a = tf.abs(self.a())
        return tf.complex(tf.exp(-a * mass), zeros)


@regist_particle("exp_com")
class ParticleExp(Particle):
    """
    .. math::
        R(m) = e^{-|a+ib| m^2}

    """

    def init_params(self):
        self.a = self.add_var("a")
        self.b = self.add_var("b")

    def get_amp(self, data, _data_c=None, **kwargs):
        mass = data["m"]
        zeros = tf.zeros_like(mass)
        a = tf.abs(self.a())
        b = self.b()
        r = -tf.complex(a, b) * tf.complex(mass * mass, zeros)
        return tf.exp(r)


@regist_decay("particle-decay")
class ParticleDecay(HelicityDecay):
    def get_ls_amp(self, data, data_p, **kwargs):
        amp = super(ParticleDecay, self).get_ls_amp(data, data_p, **kwargs)
        a = self.core
        b = self.outs[0]
        c = self.outs[1]
        mass = a.get_mass()
        width = a.get_width()
        m = data_p[a]["m"]
        if width is None:
            ret = tf.zeros_like(m)
            ret = tf.complex(ret, ret)
        elif not a.running_width:
            ret = tf.reshape(BW(m, mass, width), (-1, 1))
        else:
            q = data["|q|"]
            q0 = data["|q0|"]
            ret = []
            for i in self.get_l_list():
                bw = BWR(m, mass, width, q, q0, i, self.d)
                ret.append(tf.reshape(bw, (-1, 1)))
            ret = tf.concat(ret, axis=-1)
        return ret * amp


@regist_decay("helicity_full")
class HelicityDecayNP(HelicityDecay):
    def init_params(self):
        a = self.outs[0].spins
        b = self.outs[1].spins
        self.H = self.add_var("H", is_complex=True, shape=(len(a), len(b)))

    def get_helicity_amp(self, data, data_p, **kwargs):
        return tf.stack(self.H())


@regist_decay("helicity_full-bf")
class HelicityDecayNPbf(HelicityDecay):
    def init_params(self):
        self.d = 3.0
        a = self.outs[0].spins
        b = self.outs[1].spins
        self.H = self.add_var("H", is_complex=True, shape=(len(a), len(b)))

    def get_helicity_amp(self, data, data_p, **kwargs):
        q0 = self.get_relative_momentum(data_p, False)
        data["|q0|"] = q0
        if "|q|" in data:
            q = data["|q|"]
        else:
            q = self.get_relative_momentum(data_p, True)
            data["|q|"] = q
        bf = barrier_factor([min(self.get_l_list())], q, q0, self.d)
        H = tf.stack(self.H())
        bf = tf.cast(tf.reshape(bf, (-1, 1, 1)), H.dtype)
        return H * bf


def get_parity_term(j1, p1, j2, p2, j3, p3):
    p = p1 * p2 * p3 * (-1) ** (j1 - j2 - j3)
    return p


@regist_decay("helicity_parity")
class HelicityDecayP(HelicityDecay):
    """

    .. math::
        H_{- m1, - m2} = P_0 P_1 P_2 (-1)^{J_1 + J_2 - J_0} H_{m1, m2}

    """

    def init_params(self):
        a = self.core
        b = self.outs[0]
        c = self.outs[1]
        n_b = len(b.spins)
        n_c = len(c.spins)
        self.parity_term = get_parity_term(a.J, a.P, b.J, b.P, c.J, c.P)
        if n_b > n_c:
            self.H = self.add_var(
                "H", is_complex=True, shape=((n_b + 1) // 2, n_c)
            )
            self.part_H = 0
        else:
            self.H = self.add_var(
                "H", is_complex=True, shape=(n_b, (n_c + 1) // 2)
            )
            self.part_H = 1

    def get_helicity_amp(self, data, data_p, **kwargs):
        n_b = len(self.outs[0].spins)
        n_c = len(self.outs[1].spins)
        H_part = tf.stack(self.H())
        if self.part_H == 0:
            H = tf.concat(
                [
                    H_part,
                    self.parity_term * H_part[(n_b - 2) // 2 :: -1, ::-1],
                ],
                axis=0,
            )
        else:
            H = tf.concat(
                [
                    H_part,
                    self.parity_term * H_part[::-1, (n_c - 2) // 2 :: -1],
                ],
                axis=1,
            )
        return H


@regist_decay("gls-cpv")
class HelicityDecay(HelicityDecay):
    def init_params(self):
        self.d = 3.0
        ls = self.get_ls_list()
        self.g_ls = self.add_var(
            "g_ls", is_complex=True, shape=(len(ls),), is_cp=True
        )
        try:
            self.g_ls.set_fix_idx(fix_idx=0, fix_vals=(1.0, 0.0))
        except Exception as e:
            print(e, self, self.get_ls_list())

    def get_g_ls(self, charge=1):
        gls = self.g_ls(charge)
        if self.ls_index is None:
            return tf.stack(gls)
        # print(self, gls, self.ls_index)
        return tf.stack([gls[k] for k in self.ls_index])

    def get_ls_amp(self, data, data_p, **kwargs):
        charge = kwargs.get("all_data", {}).get("charge_conjugation", None)
        g_ls_p = self.get_g_ls(1)
        if charge is None:
            g_ls = g_ls_p
        else:
            g_ls_m = self.get_g_ls(-1)
            g_ls = tf.where((charge > 0)[:, None], g_ls_p, g_ls_m)
        # print(g_ls)
        q0 = self.get_relative_momentum2(data_p, False)
        data["|q0|2"] = q0
        if "|q|2" in data:
            q = data["|q|2"]
        else:
            q = self.get_relative_momentum2(data_p, True)
            data["|q|2"] = q
        if self.has_barrier_factor:
            bf = self.get_barrier_factor2(
                data_p[self.core]["m"], q, q0, self.d
            )
            mag = g_ls
            m_dep = mag * tf.cast(bf, mag.dtype)
        else:
            m_dep = g_ls
        return m_dep



"""
Particle and amplitude
---------------------------

    Amplitude = DecayGroup + Variable

"""

# %%
# We will use following parameters for a toy model
from tf_pwa.amp import DecayChain, DecayGroup, get_decay, get_particle

resonances = {
    "R0": {"J": 0, "P": 1, "mass": 1.0, "width": 0.07},
    "R1": {"J": 0, "P": 1, "mass": 1.0, "width": 0.07},
    "R2": {"J": 1, "P": -1, "mass": 1.225, "width": 0.08},
}

a, b, c, d = [get_particle(i, J=0, P=-1) for i in "ABCD"]
r1, r2, r3 = [get_particle(i, **resonances[i]) for i in resonances.keys()]


decay_group = DecayGroup(
    [
        DecayChain([get_decay(a, [r1, c]), get_decay(r1, [b, d])]),
        DecayChain([get_decay(a, [r2, b]), get_decay(r2, [c, d])]),
        DecayChain([get_decay(a, [r3, b]), get_decay(r3, [c, d])]),
    ]
)

# %%
# The above parts can be represented as config.yml used by ConfigLoader.
#
# We can get AmplitudeModel form decay_group and a optional Variables Managerager.
# It has parameters, so we can get and set parameters for the amplitude model
from tf_pwa.amp import AmplitudeModel
from tf_pwa.variable import VarsManager

vm = VarsManager()
amp = AmplitudeModel(decay_group, vm=vm)

print(amp.get_params())
amp.set_params(
    {
        "A->R0.CR0->B.D_total_0r": 1.0,
        "A->R1.BR1->C.D_total_0r": 1.0,
        "A->R2.BR2->C.D_total_0r": 7.0,
    }
)

# %%
# For the calculation, we generate some phase space data.

from tf_pwa.phasespace import PhaseSpaceGenerator

m_A, m_B, m_C, m_D = 1.8, 0.18, 0.18, 0.18
p1, p2, p3 = PhaseSpaceGenerator(m_A, [m_B, m_C, m_D]).generate(100000)

# %%
# and the calculate helicity angle from the data

from tf_pwa.cal_angle import cal_angle_from_momentum

data = cal_angle_from_momentum({b: p1, c: p2, d: p3}, decay_group)

# %%
# we can index mass from data as

from tf_pwa.data import data_index

m_bd = data_index(data, ("particle", "(B, D)", "m"))
# m_bc = data_index(data, ("particle", "(B, C)", "m"))
m_cd = data_index(data, ("particle", "(C, D)", "m"))

# %%
# .. note::
#     If DecayGroup do not include resonant of (B, C), the data will not include its mass too.
#     We can use different DecayGroup for cal_angle and AmplitudeModel
#     when they have the same initial and final particle.
#
# The amplitde square is calculated by amplitude model simply as

amp_s2 = amp(data)

# %%
# Now by using matplotlib we can get the Dalitz plot as

import matplotlib.pyplot as plt

plt.clf()
plt.hist2d(
    m_bd.numpy() ** 2,
    m_cd.numpy() ** 2,
    weights=amp_s2.numpy(),
    bins=60,
    cmin=1,
    cmap="jet",
)
plt.colorbar()
plt.xlabel("$m^2(BD)$")
plt.ylabel("$m^2(CD)$")
plt.show()



"""
This module provides functions that implements user-friendly interface to the functions and methods in other modules.
It acts like a synthesis of all the other modules of their own physical purposes.
In general, users only need to import functions in this module to implement their physical analysis instead of
going into every modules. There are some example files where you can figure out how it is used.
"""
import os
import time
import warnings

import matplotlib.pyplot as plt
import numpy as np
import tensorflow as tf
from scipy.stats import norm as Norm

from .cal_angle import cal_angle_from_momentum, prepare_data_from_decay
from .data import data_to_tensor, load_dat_file, split_generator
from .fit import fit_minuit, fit_multinest, fit_scipy
from .fitfractions import cal_fitfractions, cal_fitfractions_no_grad
from .phasespace import PhaseSpaceGenerator
from .significance import significance
from .utils import check_positive_definite, error_print, std_periodic_var


def fit_fractions(
    amp, mcdata, inv_he=None, params=None, batch=25000, res=None
):
    """
    This function calculate fit fractions of the resonances as well as their coherent pairs. It imports
    ``cal_fitfractions`` and ``cal_fitfractions_no_grad`` from module **tf_pwa.fitfractions**.

    .. math:: FF_{i} = \\frac{\\int |A_i|^2 d\\Omega }{ \\int |\\sum_{i}A_i|^2 d\\Omega }\\approx \\frac{\\sum |A_i|^2 }{\\sum|\\sum_{i} A_{i}|^2}

    gradients???:

    .. math:: FF_{i,j} = \\frac{\\int 2Re(A_i A_j*) d\\Omega }{ \\int |\\sum_{i}A_i|^2 d\\Omega } = \\frac{\\int |A_i +A_j|^2  d\\Omega }{ \\int |\\sum_{i}A_i|^2 d\\Omega } - FF_{i} - FF_{j}

    hessians:

    .. math:: \\frac{\\partial }{\\partial \\theta_i }\\frac{f(\\theta_i)}{g(\\theta_i)} = \\frac{\\partial f(\\theta_i)}{\\partial \\theta_i} \\frac{1}{g(\\theta_i)} - \\frac{\\partial g(\\theta_i)}{\\partial \\theta_i} \\frac{f(\\theta_i)}{g^2(\\theta_i)}

    :param amp: Amplitude object.
    :param mcdata: MCdata array.
    :param inv_he: The inverse of Hessian matrix. If it's not given, the errors will not be calculated.
    :return frac: Dictionary of fit fractions for each resonance.
    :return err_frac: Dictionary of their errors. If ``inv_he`` is ``None``, it will be a dictionary of ``None``.
    """
    if params is None:
        params = {}
    err_frac = {}
    with amp.temp_params(params):
        frac, grad = cal_fitfractions(amp, mcdata, res=res, batch=batch)
    if inv_he is not None:
        for i in frac:
            err_frac[i] = np.sqrt(np.dot(np.dot(inv_he, grad[i]), grad[i]))
    return frac, err_frac


def corr_coef_matrix(err_mtx):
    """
    This function obtains correlation coefficients matrix of all trainable variables from `*.npy` file.

    :param err_mtx: Array or string (name of the npy file).
    :return: Numpy 2-d array. The correlation coefficient matrix.
    """
    if type(err_mtx) is str:
        err_mtx = np.load(err_mtx)
    err = np.sqrt(np.fabs(err_mtx.diagonal()))
    diag_mtx = np.diag(1 / err)
    tmp_mtx = np.matmul(diag_mtx, err_mtx)
    cc_mtx = np.matmul(tmp_mtx, diag_mtx)
    return cc_mtx


'''
def calPWratio(params, POLAR=True):
    """
    This function calculates the ratio of different partial waves in a certain resonance using the input values
    of fitting parameters. It is useful when user check if one partial wave is too small compared to the other partial
    waves, in which case, the fitting result may be not reliable since it has potential to give nuisance
    degree of freedom. (WIP)

    :param params: Dictionary of values indexed by the name of the fitting parameters
    :param POLAR: Boolean. Whether the parameters are defined in the polar coordinate or the Cartesian coordinate.
    :return: None. But the function will print the ratios.
    """
    dtype = "float64"
    w_bkg = 0.768331
    # set_gpu_mem_growth()
    tf.keras.backend.set_floatx(dtype)
    config_list = load_config_file("Resonances")
    a = Model(config_list, w_bkg, kwargs={"polar": POLAR})

    args_name = []
    for i in a.Amp.trainable_variables:
        args_name.append(i.name)
    # a.Amp.polar=True

    a.set_params(params)
    if not POLAR:  # if final_params.json is not in polar coordinates
        i = 0
        for v in args_name:
            if len(v) > 15:
                if i % 2 == 0:
                    tmp_name = v
                    tmp_val = params[v]
                else:
                    params[tmp_name] = np.sqrt(tmp_val ** 2 + params[v] ** 2)
                    params[v] = np.arctan2(params[v], tmp_val)
            i += 1
        a.set_params(params)
    fname = [["data/data4600_new.dat", "data/Dst0_data4600_new.dat"],
             ["data/bg4600_new.dat", "data/Dst0_bg4600_new.dat"],
             ["data/PHSP4600_new.dat", "data/Dst0_PHSP4600_new.dat"]
             ]
    tname = ["data", "bg", "PHSP"]
    data_np = {}
    for i in range(3):
        data_np[tname[i]] = cal_ang_file(fname[i][0], dtype)

    def load_data(name):
        dat = []
        tmp = flatten_np_data(data_np[name])
        for i in param_list:
            tmp_data = tf.Variable(tmp[i], name=i, dtype=dtype)
            dat.append(tmp_data)
        return dat

    data = load_data("data")
    bg = load_data("bg")
    mcdata = load_data("PHSP")
    data_cache = a.Amp.cache_data(*data)
    bg_cache = a.Amp.cache_data(*bg)
    mcdata_cache = a.Amp.cache_data(*mcdata)
    total = a.Amp(mcdata_cache, cached=True)
    a_sig = {}

    # res_list = [[i] for i in config_list]
    res_list = [
        ["Zc_4025"],
        ["Zc_4160"],
        ["D1_2420", "D1_2420p"],
        ["D1_2430", "D1_2430p"],
        ["D2_2460", "D2_2460p"],
    ]

    config_res = [part_config(config_list, i) for i in res_list]
    PWamp = {}
    for i in range(len(res_list)):
        name = res_list[i]
        if isinstance(name, list):
            if len(name) > 1:
                name = reduce(lambda x, y: "{}+{}".format(x, y), res_list[i])
            else:
                name = name[0]
        a_sig[i] = Model(config_res[i], w_bkg, kwargs={"polar": POLAR})
        p_list = [[], []]
        for p in a_sig[i].get_params():
            if p[-3] == 'r' and len(p) > 15:
                if p[8] == 'd':
                    p_list[1].append(p)
                else:
                    p_list[0].append(p)
        first = True
        for p in p_list[0]:
            a_sig[i].set_params(params)
            for q in p_list[0]:
                a_sig[i].set_params({q: 0})
            a_sig[i].set_params({p: params[p]})
            if first:
                norm = a_sig[i].Amp(mcdata_cache, cached=True).numpy().sum()
                print(p[:-3], "\t", 1.0)
                first = False
            else:
                print(p[:-3], "\t", a_sig[i].Amp(mcdata_cache, cached=True).numpy().sum() / norm)
        first = True
        for p in p_list[1]:
            a_sig[i].set_params(params)
            for q in p_list[1]:
                a_sig[i].set_params({q: 0})
            a_sig[i].set_params({p: params[p]})
            if first:
                norm = a_sig[i].Amp(mcdata_cache, cached=True).numpy().sum()
                print(p[:-3], "\t", 1.0)
                first = False
            else:
                print(p[:-3], "\t", a_sig[i].Amp(mcdata_cache, cached=True).numpy().sum() / norm)
        print()
        # print(a_sig[i].get_params())
        # a_weight[i] = a_sig[i].Amp(mcdata_cache,cached=True).numpy()
        # PWamp[name] = a_weight[i].sum()/(n_data - w_bkg*n_bg)
'''


def force_pos_def_minuit2(inv_he):
    """
    force positive defined of error matrix

        from minuit2 https://github.com/root-project/root/blob/master/math/minuit2/sec/MnPosDef.cxx
    """
    inv_he = np.array(inv_he)
    finfo = np.info(inv_he.dtype)
    # print(finfo, inv_he.dtype)
    if finfo is not None:
        eps = finfo.eps
    else:
        eps = 2.220446049250313e-16  # float64
    eps2 = 2 * np.sqrt(eps)

    diag = np.diagonal(inv_he)
    diag_i = np.ones_like(diag)
    dgmin = np.min(diag)
    if dgmin > 0:
        return inv_he
    epspdf = max(1e-6, eps2)
    dg = 0.5 + epspdf - dgmin
    inv_he += np.diag(diag_i * dg)
    warnings.warn("Added to diagonal of Error Matrix a value {}".format(dgmin))
    diag2 = np.diagonal(inv_he)
    diag2 = np.where(diag2 < 0, 1, diag2)
    e = 1 / np.sqrt(diag2)
    p = inv_he.copy() / diag[:, None] / diag[None, :]
    p = np.triu(p, 0)
    e, v = np.linalg.eig(p)
    pmin = np.min(e)
    pmax = max(np.max(np.abs(e)), 1)
    if pmin > epspdf * pmax:
        return inv_he

    padd = 0.001 * pmax - pmin
    inv_he += np.diag(np.diag(inv_he) * padd)
    warnings.warn(
        "Matrix forced pos-def by adding to diagonal {}".format(padd)
    )
    return inv_he


def force_pos_def(h):
    """
    from pricession lost hessian matrix
    eigen value is small

    dot(H, v[:,i] = e[i] v[:,i]
    dot(H, v)[:,i] = e[i] v[:,i]
    dot(inv(v), dot(H, v)) = diag(e)
    H = dot(v, dot(diag(e), inv(v))

    """
    e, v = np.linalg.eig(h)
    inv_he = np.linalg.pinv(h)
    e_min = np.min(e)
    e_max = np.max(np.abs(e))
    if e_min > 0:
        return inv_he
    if e_min > -0.00001 * e_max:  # pricession cause small nagtive eigen value
        idx = np.argmin(e)
        e_new = np.where(e < 0, e - e_min * 1.1, e)
        warnings.warn(
            "Matrix forced pos-def by adding {} to eigen value".format(
                -e_min * 1.1
            )
        )
        h = np.dot(v, np.dot(np.diag(e_new), np.linalg.inv(v)))
        inv_he = np.linalg.inv(h)
        return inv_he
    return force_pos_def_minuit2(inv_he)


def cal_hesse_error(
    fcn, params={}, check_posi_def=True, force_pos=True, save_npy=True
):
    """
    This function calculates the errors of all trainable variables.
    The errors are given by the square root of the diagonal of the inverse Hessian matrix.

    :param model: Model.
    :return hesse_error: List of errors.
    :return inv_he: The inverse Hessian matrix.
    """
    t = time.time()
    nll, g, h = fcn.nll_grad_hessian(
        params
    )  # data_w,mcdata,weight=weights,batch=50000)
    print("Time for calculating errors:", time.time() - t)
    h = h.numpy()
    inv_he = np.linalg.inv(h)
    if check_posi_def and check_positive_definite(inv_he):
        inv_he = np.linalg.inv(h)
    else:
        inv_he = np.linalg.pinv(h)
        if force_pos:
            inv_he = force_pos_def(h)

    if save_npy:
        np.save("error_matrix.npy", inv_he)
    diag_he = inv_he.diagonal()
    hesse_error = np.sqrt(np.fabs(diag_he)).tolist()
    return hesse_error, inv_he


def cal_hesse_correct(fcn, params={}, corr_params={}, force_pos=True):
    t = time.time()
    nll, g, h = fcn.nll_grad_hessian(params)
    # data_w,mcdata,weight=weights,batch=50000)
    h = h.numpy()
    var_names = fcn.vm.trainable_vars
    params0 = fcn.get_params()
    x0 = np.array([params0[k] for k in var_names])
    idxs = [var_names.index(k) for k in corr_params]
    # calculate hessian matrix for special params
    _epsilon = 1e-3
    _epsilon2 = _epsilon * 2
    x = x0.copy()
    for i in idxs:
        for j in range(len(var_names)):
            if j in idxs and i > j:
                continue
            if i == j:
                x[i] += 2 * _epsilon
                nll_pp = fcn(x)
                x[i] -= _epsilon
                nll_pm = fcn(x)
                x[i] -= 2 * _epsilon
                nll_mp = fcn(x)
                x[i] -= _epsilon
                nll_mm = fcn(x)
                x[i] += 2 * _epsilon
                gp = (nll_pp - nll_mp) / 3 / _epsilon
                gm = (nll_mp - nll_mm) / 3 / _epsilon
                new_hi = (gp - gm) / _epsilon
                h[i, i] = new_hi
            else:
                x[i] += _epsilon
                x[j] += _epsilon
                nll_pp = fcn(x)
                x[j] -= 2 * _epsilon
                nll_pm = fcn(x)
                x[j] += _epsilon
                x[i] -= 2 * _epsilon
                x[j] += _epsilon
                nll_mp = fcn(x)
                x[j] -= 2 * _epsilon
                nll_mm = fcn(x)
                x[j] += _epsilon
                x[i] += _epsilon
                gp = (nll_pp - nll_pm) / _epsilon2
                gm = (nll_mp - nll_mm) / _epsilon2
                new_hi = (gp - gm) / _epsilon2
                h[i, j] = new_hi
                h[j, i] = new_hi
    print("Time for calculating errors:", time.time() - t)
    return h


def num_hess_inv_3point(fcn, params={}, _epsilon=5e-4):
    """
    This function calculates the errors of all trainable variables.
    The errors are given by the square root of the diagonal of the inverse Hessian matrix.

    :param model: Model.
    :return hesse_error: List of errors.
    :return inv_he: The inverse Hessian matrix.
    """
    var_names = fcn.vm.trainable_vars
    old_params = fcn.get_params()
    fcn(params)
    f_g = fcn.vm.trans_fcn_grad(fcn.nll_grad)
    hess = []
    x0 = np.array(fcn.vm.get_all_val(True))
    for i in range(len(var_names)):
        x0[i] += _epsilon
        _, g1 = f_g(x0)
        x0[i] -= 2 * _epsilon
        _, g2 = f_g(x0)
        x0[i] += _epsilon
        hess.append([(i - j) / 2 / _epsilon for i, j in zip(g1, g2)])
    hess_inv = np.linalg.inv(hess)
    fcn(old_params)
    hess_inv = fcn.vm.trans_error_matrix(hess_inv, x0)
    return hess_inv


def gen_data(
    amp,
    Ndata,
    mcfile,
    Nbg=0,
    wbg=0,
    Poisson_fluc=False,
    bgfile=None,
    genfile=None,
    particles=None,
):
    """
    This function is used to generate toy data according to an amplitude model.

    :param amp: AmplitudeModel???
    :param particles: List of final particles
    :param Ndata: Integer. Number of data
    :param mcfile: String. The MC sample file used to generate signal data.
    :param Nbg: Integer. Number of background. By default it's 0.
    :param wbg: Float. Weight of background. By default it's 0.
    :param Poisson_fluc: Boolean. If it's ``True``, The number of data will be decided by a Poisson distribution around the given value.
    :param bgfile: String. The background sample file used to generate a certain number of background data.
    :param genfile: String. The file to store the generated toy.
    :return: tensorflow.Tensor. The generated toy data.
    """
    Nbg = round(wbg * Nbg)
    Nmc = Ndata - Nbg  # 8065-3445*0.768331
    if Poisson_fluc:  # Poisson
        Nmc = np.random.poisson(Nmc)
        Nbg = np.random.poisson(Nbg)
    print("data:", Nmc + Nbg, ", sig:", Nmc, ", bkg:", Nbg)
    dtype = "float64"
    if not particles:
        particles = sorted(amp.decay_group.outs)
    phsp = prepare_data_from_decay(mcfile, amp.decay_group, dtype=dtype)
    phsp = data_to_tensor(phsp)
    ampsq = amp(phsp)
    ampsq_max = tf.reduce_max(ampsq).numpy()
    Nsample = ampsq.__len__()
    n = 0
    idx_list = []

    while n < Nmc:
        uni_rdm = tf.random.uniform(
            [Nsample], minval=0, maxval=ampsq_max, dtype=dtype
        )
        list_rdm = tf.random.uniform([Nsample], dtype=tf.int64, maxval=Nsample)
        j = 0
        mask = tf.boolean_mask(list_rdm, tf.gather(ampsq, list_rdm) > uni_rdm)
        idx_list.append(mask)
        n += mask.shape[0]
        # print(mask)
        # for i in list_rdm:
        # if ampsq[i] > uni_rdm[j]:
        # idx_list.append(i)
        # n += 1
        # j += 1
        # if n == Nmc:
        # break
    idx_list = tf.concat(idx_list, axis=0).numpy()[:Nmc]

    data_tmp = load_dat_file(mcfile, particles, dtype)
    for i in particles:
        data_tmp[i] = np.array(data_tmp[i])[idx_list]
    data_gen = []

    if Nbg:
        bg_tmp = load_dat_file(bgfile, particles, dtype)
        bg_idx = tf.random.uniform(
            [Nbg], dtype=tf.int64, maxval=len(bg_tmp[particles[0]])
        )  # np.random.randint(len(bg),size=Nbg)
        bg_idx = tf.stack(bg_idx).numpy()
        for i in particles:
            tmp = bg_tmp[i][bg_idx]
            data_tmp[i] = np.append(data_tmp[i], tmp, axis=0)
            data_gen.append(data_tmp[i])
    else:
        for i in particles:
            data_gen.append(data_tmp[i])

    data_gen = np.transpose(data_gen, [1, 0, 2])
    np.random.shuffle(data_gen)
    data_gen = data_gen.reshape(-1, 4)

    if isinstance(genfile, str):
        np.savetxt(genfile, data_gen)
    # data = prepare_data_from_decay(genfile, amp.decay_group, particles=particles, dtype=dtype)

    momenta = {}
    Npar = len(particles)
    for p in range(Npar):
        momenta[particles[p]] = data_gen[p::Npar]
    data = cal_angle_from_momentum(momenta, amp.decay_group)
    return data_to_tensor(data)


def gen_mc(mother, daughters, number, outfile=None):
    """
    This function generates phase-space MC data (without considering the effect of detector performance). It imports
    ``PhaseSpaceGenerator`` from module **tf_pwa.phasespace**.

    :param mother: Float. The invariant mass of the mother particle.
    :param daughters: List of float. The invariant masses of the daughter particles.
    :param number: Integer. The number of MC data generated.
    :param outfile: String. The file to store the generated MC.
    :return: Numpy array. The generated MC data.
    """
    # 4.59925172, [2.00698, 2.01028, 0.13957] DBC: D*0 D*- pi+
    phsp = PhaseSpaceGenerator(mother, daughters)
    flat_mc_data = phsp.generate(number)
    pf = []
    for i in range(len(daughters)):
        pf.append(flat_mc_data[i])
    pf = np.transpose(pf, (1, 0, 2)).reshape((-1, 4))
    if isinstance(outfile, str):
        np.savetxt(outfile, pf)  # 一个不包含探测器效率的MC样本
    return pf


def fit(Use="scipy", **kwargs):
    """
    Fit the amplitude model using ``scipy``, ``iminuit`` or ``pymultinest``. It imports
    ``fit_scipy``, ``fit_minuit``, ``fit_multinest`` from module **tf_pwa.fit**.

    :param Use: String. If it's ``"scipy"``, it will call ``fit_scipy``; if it's ``"minuit"``, it will call ``fit_minuit``; if it's ``"multinest"``, it will call ``fit_multinest``.
    :param kwargs: The arguments will be passed to the three functions above.

    For ``fit_scipy``

    :param fcn: FCN object to be minimized.
    :param method: String. Options in ``scipy.optimize``. For now, it implements interface to such as "BFGS", "L-BFGS-B", "basinhopping".
    :param bounds_dict: Dictionary of boundary constrain to variables.
    :param kwargs: Other arguments passed on to ``scipy.optimize`` functions.
    :return: FitResult object, List of NLLs, List of point arrays.

    For ``fit_minuit``

    :param fcn: FCN object to be minimized.
    :param bounds_dict: Dictionary of boundary constrain to variables.
    :param hesse: Boolean. Whether to call ``hesse()`` after ``migrad()``. It's ``True`` by default.
    :param minos: Boolean. Whether to call ``minos()`` after ``hesse()``. It's ``False`` by default.
    :return: Minuit object

    For ``fit_multinest`` (WIP)

    :param fcn: FCN object to be minimized.
    """
    method = kwargs.get("method", "BFGS")
    if method in ["minuit"]:
        Use = "minuit"
    if Use == "scipy":
        ret = fit_scipy(**kwargs)
    elif Use == "minuit":
        ret = fit_minuit(**kwargs)
    # elif Use == "multinest":
    #    ret = fit_multinest(**kwargs)
    else:
        raise Exception("Unknown fit tool {}".format(Use))
    return ret


def cal_significance(nll1, nll2, ndf):
    """
    This function calculates the statistical significance.

    :param nll1: Float. NLL of the first PDF.
    :param nll2:  Float. NLL of the second PDF.
    :param ndf: The difference of the degrees of freedom of the two PDF.
    :return: Float. The statistical significance
    """
    sigma = significance(nll1, nll2, ndf)
    return sigma


def plot_pull(data, name, nbins=20, norm=False, value=None, error=None):
    """
    This function is used to plot the pull for a data sample.

    :param data: List
    :param name: String. Name of the sample
    :param nbins: Integer. Number of bins in the histogram
    :param norm: Boolean. Whether to normalize the histogram
    :param value: Float. Mean value in normalization
    :param error: Float or list. Sigma value(s) in normalization
    :return: The fitted mu, sigma, as well as their errors
    """
    data = np.array(data)
    if norm:
        if value is None or error is None:
            raise Exception("Need value or error for normed pull!")
        data = (data - value) / error

    _, bins, _ = plt.hist(data, nbins, density=True, alpha=0.6)
    bins = np.linspace(bins[0], bins[-1], 30)

    try:
        from iminuit import Minuit

        print("Using Minuit to fit the histogram")

        def fitNormHist(data):
            def nll(mu, sigma):
                def normpdf(x, mu, sigma):
                    return (
                        np.exp(-(x - mu) * (x - mu) / 2 / sigma / sigma)
                        / np.sqrt(2 * np.pi)
                        / sigma
                    )

                return -np.sum(np.log(normpdf(data, mu, sigma)))

            m = Minuit(
                nll, mu=0, sigma=1, error_mu=0.1, error_sigma=0.1, errordef=0.5
            )
            m.migrad()
            m.hesse()
            return m

        m = fitNormHist(data)
        mu, sigma = m.values["mu"], m.values["sigma"]
        err_mu, err_sigma = m.errors["mu"], m.errors["sigma"]

    except Exception as e:
        if type(e) is ModuleNotFoundError:
            print(
                "No Minuit installed. Using scipy.optimize to fit the histogram"
            )
            from scipy.optimize import minimize

            def fitNormHist(data):
                def nll(mu_sigma):
                    def normpdf(x, mu_sigma):
                        return (
                            np.exp(
                                -(x - mu_sigma[0])
                                * (x - mu_sigma[0])
                                / 2
                                / mu_sigma[1]
                                / mu_sigma[1]
                            )
                            / np.sqrt(2 * np.pi)
                            / mu_sigma[1]
                        )

                    return -np.sum(np.log(normpdf(data, mu_sigma)))

                s = minimize(nll, [0, 1])
                return s

            s = fitNormHist(data)
            mu, sigma = s.x
            err_mu, err_sigma = np.sqrt(np.diagonal(s.hess_inv))
        else:
            raise e

    y = Norm.pdf(bins, mu, sigma)
    plt.plot(bins, y, "r-")
    plt.xlabel(name)
    plt.title(
        "mu = "
        + error_print(mu, err_mu)
        + "; sigma = "
        + error_print(sigma, err_sigma)
    )
    plt.savefig("fig/" + name + "_pull.png")
    plt.clf()
    return mu, sigma, err_mu, err_sigma


# def likelihood_profile(var_name, start=None, end=None, step=None, values=None, errors=None, mode="bothsides"):
# if start == None or end == None:
#    x_mean = values[var_name]
#    x_sigma = errors[var_name]
#    start = x_mean - 10 * x_sigma
#    end = x_mean + 10 * x_sigma
# else:
#    x_mean = (end + start) / 2
# if step == None:
#    step = (end - start) / 100
# if mode == "bothsides":
#    x1 = np.arange(x_mean, start, -step)
#    x2 = np.arange(x_mean, end, step)
# elif mode == "back&forth":
#    x1 = np.arange(start, end, step)
#    x2 = x1[::-1]


def likelihood_profile(m, var_names, bins=20, minos=True):
    """
    Calculate the likelihood profile for a variable.

    :param m: Minuit object
    :param var_names: Either a string or a list of strings
    :param bins: Integer
    :param minos: Boolean. If it's ``False``, the function will call ``Minuit.profile()`` to derive the 1-d scan of **var_names**; if it's ``True``, the function will call ``Minuit.mnprofile()`` to derive the likelihood profile, which is much more time-consuming.
    :return: Dictionary indexed by **var_names**. It contains the return of either ``Minuit.mnprofile()`` or ``Minuit.profile()``.
    """
    if isinstance(var_names, str):
        var_names = [var_names]
    lklpf = {}
    for var in var_names:
        if minos:
            x, y, t = m.mnprofile(var, bins=bins)
            lklpf[var] = [x, y, t]
        else:
            x, y = m.profile(var, bins=bins)
            lklpf[var] = [x, y]
    return lklpf


def compare_result(
    value1,
    value2,
    error1,
    error2=None,
    figname=None,
    yrange=None,
    periodic_vars=None,
):
    """
    Compare two groups of fitting results. If only one error is provided,
    the figure is :math:`\\frac{\\mu_1-\\mu_2}{\\sigma_1}`;
    if both errors are provided, the figure is :math:`\\frac{\\mu_1-\\mu_2}{\\sqrt{\\sigma_1^2+\\sigma_2^2}}`.

    :param value1: Dictionary
    :param value2: Dictionary
    :param error1: Dictionary
    :param error2: Dictionary. By default it's ``None``.
    :param figname: String. The output file
    :param yrange: Float. If it's not given, there is no y-axis limit in the figure.
    :param periodic_vars: List of strings. The periodic variables.
    :return: Dictionary of quality figure of each variable.
    """
    if periodic_vars is None:
        periodic_vars = []
    diff_dict = {}
    if error2:
        for v in error1:
            if v in periodic_vars:
                diff = value1[v] - std_periodic_var(value2[v], value1[v])
            else:
                diff = value1[v] - value2[v]
            sigma = np.sqrt(error1[v] ** 2 + error2[v] ** 2)
            diff_dict[v] = diff / sigma
    else:
        for v in error1:
            if v in periodic_vars:
                diff = value1[v] - std_periodic_var(value2[v], value1[v])
            else:
                diff = value1[v] - value2[v]
            diff_dict[v] = diff / error1[v]
    if figname:
        arr = []
        if yrange:
            for v in diff_dict:
                if np.abs(diff_dict[v]) > yrange:
                    print(
                        "{0} out of yrange, which is {1}.".format(
                            v, diff_dict[v]
                        )
                    )
                    arr.append(np.sign(diff_dict[v]) * yrange)
                else:
                    arr.append(diff_dict[v])
            # pylint: disable=invalid-unary-operand-type
            plt.ylim(-yrange, yrange)
        else:
            for v in diff_dict:
                arr.append(diff_dict[v])
        arr_x = np.arange(len(arr))
        plt.scatter(arr_x, arr)
        plt.xlabel("parameter index")
        plt.ylabel("sigma")
        plt.title(figname)
        plt.savefig(figname + ".png")
        plt.clf()
    return diff_dict



"""
module for describing data process.

All data structure is describing as nested combination of `dict` or `list` for `ndarray`.
Data process is a translation from data structure to another data structure or typical `ndarray`.
Data cache can be implemented based on the dynamic features of `list` and `dict`.

The full data structure is

.. code::

    {
    "particle":{
        "A":{"p":...,"m":...}
        ...
    },
    "decay":[
        {
        "A->R1+B": {
            "R1": {
            "ang":  {
                "alpha":[...],
                "beta": [...],
                "gamma": [...]
            },
            "z": [[x1,y1,z1],...],
            "x": [[x2,y2,z2],...]
            },
            "B" : {...}
        },
        "R->C+D": {
            "C": {
            ...,
            "aligned_angle":{
                "alpha":[...],
                "beta":[...],
                "gamma":[...]
            }
            },
            "D": {...}
        },
        },
        {
        "A->R2+C": {...},
        "R2->B+D": {...}
        },
        ...
    ],
    "weight": [...]
    }

"""

import random
from pprint import pprint

import numpy as np

from .config import get_config
from .tensorflow_wrapper import tf, tf_version

# import tensorflow as tf
# from pysnooper import  snoop


try:
    from collections.abc import Iterable
except ImportError:  # python version < 3.7
    from collections import Iterable


def set_random_seed(seed):
    """
    set random seed for random, numpy and tensorflow
    """
    np.random.seed(seed)
    tf.random.set_seed(seed)
    random.seed(seed)


def load_dat_file(
    fnames, particles, dtype=None, split=None, order=None, _force_list=False
):
    """
    Load ``*.dat`` file(s) of 4-momenta of the final particles.

    :param fnames: String or list of strings. File names.
    :param particles: List of Particle. Final particles.
    :param dtype: Data type.
    :param split: sizes of each splited dat files
    :param order: transpose order

    :return: Dictionary of data indexed by Particle.
    """
    n = len(particles)
    if dtype is None:
        dtype = get_config("dtype")

    if isinstance(fnames, str):
        fnames = [fnames]
    elif isinstance(fnames, Iterable):
        fnames = list(fnames)
    else:
        raise TypeError("fnames must be string or list of strings")

    datas = []
    sizes = []
    for fname in fnames:
        if fname.endswith(".npz"):
            data = np.load(fname)["arr_0"]
        elif fname.endswith(".npy"):
            data = np.load(fname)
        else:
            data = np.loadtxt(fname, dtype=dtype)
        data = np.reshape(data, (-1, 4))
        sizes.append(data.shape[0])
        datas.append(data)

    if split is None:
        n_total = sum(sizes)
        if n_total % n != 0:
            raise ValueError("number of data find {}/{}".format(n_total, n))
        n_data = n_total // n
        split = [size // n_data for size in sizes]

    if order is None:
        order = (1, 0, 2)

    ret = {}
    idx = 0
    for size, data in zip(split, datas):
        data_1 = data.reshape((-1, size, 4))
        data_2 = data_1.transpose(order)
        for i in data_2:
            part = particles[idx]
            ret[part] = i
            idx += 1

    return ret


def save_data(file_name, obj, **kwargs):
    """Save structured data to files. The arguments will be passed to ``numpy.save()``."""
    return np.save(file_name, obj, **kwargs)


def save_dataz(file_name, obj, **kwargs):
    """Save compressed structured data to files. The arguments will be passed to ``numpy.save()``."""
    return np.savez(file_name, obj, **kwargs)


def load_data(file_name, **kwargs):
    """Load data file from save_data. The arguments will be passed to ``numpy.load()``."""
    if "allow_pickle" not in kwargs:
        kwargs["allow_pickle"] = True
    data = np.load(file_name, **kwargs)
    try:
        return data["arr_0"].item()
    except IndexError:
        try:
            return data.item()
        except ValueError:
            return data


def _data_split(dat, batch_size, axis=0):
    data_size = dat.shape[axis]
    if axis == 0:
        for i in range(0, data_size, batch_size):
            yield dat[i : min(i + batch_size, data_size)]
    elif axis == -1:
        for i in range(0, data_size, batch_size):
            yield dat[..., i : min(i + batch_size, data_size)]
    else:
        raise Exception("unsupported axis: {}".format(axis))


def data_generator(data, fun=_data_split, args=(), kwargs=None, MAX_ITER=1000):
    """Data generator: call ``fun`` to each ``data`` as a generator. The extra arguments will be passed to ``fun``."""
    kwargs = kwargs if kwargs is not None else {}

    def _gen(dat):
        if isinstance(dat, dict):
            if not dat:
                for i in range(MAX_ITER):
                    yield {}
            ks, vs = [], []
            for k, v in dat.items():
                ks.append(k)
                vs.append(_gen(v))
            for s_data in zip(*vs):
                yield type(dat)(zip(ks, s_data))
        elif isinstance(dat, list):
            if not dat:
                for i in range(MAX_ITER):
                    yield []
            vs = []
            for v in dat:
                vs.append(_gen(v))
            for s_data in zip(*vs):
                yield list(s_data)
        elif isinstance(dat, tuple):
            vs = []
            for v in dat:
                vs.append(_gen(v))
            for s_data in zip(*vs):
                yield s_data
        else:
            for i in fun(dat, *args, **kwargs):
                yield i

    return _gen(data)


def data_split(data, batch_size, axis=0):
    """
    Split ``data`` for ``batch_size`` each in ``axis``.

    :param data: structured data
    :param batch_size: Integer, data size for each split data
    :param axis: Integer, axis for split, [option]
    :return: a generator for split data

    >>> data = {"a": [np.array([1.0, 2.0]), np.array([3.0, 4.0])], "b": {"c": np.array([5.0, 6.0])}, "d": [], "e": {}}
    >>> for i, data_i in enumerate(data_split(data, 1)):
    ...     print(i, data_to_numpy(data_i))
    ...
    0 {'a': [array([1.]), array([3.])], 'b': {'c': array([5.])}, 'd': [], 'e': {}}
    1 {'a': [array([2.]), array([4.])], 'b': {'c': array([6.])}, 'd': [], 'e': {}}

    """
    return data_generator(
        data, fun=_data_split, args=(batch_size,), kwargs={"axis": axis}
    )


split_generator = data_split


def data_map(data, fun, args=(), kwargs=None):
    """Apply fun for each data. It returns the same structure."""
    kwargs = kwargs if kwargs is not None else {}
    if isinstance(data, dict):
        return type(data)(
            {k: data_map(v, fun, args, kwargs) for k, v in data.items()}
        )
    if isinstance(data, list):
        return [data_map(data_i, fun, args, kwargs) for data_i in data]
    if isinstance(data, tuple):
        return tuple([data_map(data_i, fun, args, kwargs) for data_i in data])
    return fun(data, *args, **kwargs)


def data_struct(data):
    """get the structure of data, keys and shape"""
    if isinstance(data, dict):
        return type(data)({k: data_struct(v) for k, v in data.items()})
    if isinstance(data, list):
        return [data_struct(data_i) for data_i in data]
    if isinstance(data, tuple):
        return tuple([data_struct(data_i) for data_i in data])
    if hasattr(data, "shape"):
        return tuple(data.shape)
    return data


def data_mask(data, select):
    """
    This function using boolean mask to select data.

    :param data: data to select
    :param select: 1-d boolean array for selection
    :return: data after selection
    """
    ret = data_map(data, tf.boolean_mask, args=(select,))
    return ret


def data_cut(data, expr, var_map=None):
    """cut data with boolean expression

    :param data: data need to cut
    :param expr: cut expression
    :param var_map: variable map between parameters in expr and data, [option]

    :return: data after being cut,
    """
    var_map = var_map if isinstance(var_map, dict) else {}
    import sympy as sym

    expr_s = sym.sympify(expr)
    params = tuple(expr_s.free_symbols)
    args = [data_index(data, var_map.get(i.name, i.name)) for i in params]
    expr_f = sym.lambdify(params, expr, "tensorflow")
    mask = expr_f(*args)
    return data_mask(data, mask)


def data_merge(*data, axis=0):
    """This function merges data with the same structure."""
    assert len(data) > 0
    if isinstance(data[0], dict):
        assert all([isinstance(i, dict) for i in data]), "not all type same"
        all_idx = [set(list(i)) for i in data]
        idx = set.intersection(*all_idx)
        return type(data[0])(
            {i: data_merge(*[data_i[i] for data_i in data]) for i in idx}
        )
    if isinstance(data[0], list):
        assert all([isinstance(i, list) for i in data]), "not all type same"
        return [data_merge(*data_i) for data_i in zip(*data)]
    if isinstance(data[0], tuple):
        assert all([isinstance(i, tuple) for i in data]), "not all type same"
        return tuple([data_merge(*data_i) for data_i in zip(*data)])
    m_data = tf.concat(data, axis=axis)
    return m_data


def data_shape(data, axis=0, all_list=False):
    """
    Get data size.

    :param data: Data array
    :param axis: Integer. ???
    :param all_list: Boolean. ???
    :return:
    """

    def flatten(dat):
        ret = []

        def data_list(dat1):
            if hasattr(dat1, "shape"):
                ret.append(dat1.shape)
            else:
                ret.append(())

        data_map(dat, data_list)
        return ret

    shapes = flatten(data)
    if all_list:
        return shapes
    ret = shapes[0][axis]
    if tf_version < 2:
        return ret.value
    return ret


def data_to_numpy(dat):
    """Convert Tensor data to ``numpy.ndarray``."""

    def to_numpy(data):
        if hasattr(data, "numpy"):
            return data.numpy()
        return data

    dat = data_map(dat, to_numpy)
    return dat


def data_to_tensor(dat):
    """convert data to ``tensorflow.Tensor``."""

    def to_tensor(data):
        return tf.convert_to_tensor(data)

    dat = data_map(dat, to_tensor)
    return dat


def flatten_dict_data(data, fun="{}/{}".format):
    """Flatten data as dict with structure named as ``fun``."""

    def dict_gen(dat):
        return dat.items()

    def list_gen(dat):
        return enumerate(dat)

    if isinstance(data, (dict, list, tuple)):
        ret = {}
        gen_1 = dict_gen if isinstance(data, dict) else list_gen
        for i, data_i in gen_1(data):
            tmp = flatten_dict_data(data_i)
            if isinstance(tmp, (dict, list, tuple)):
                gen_2 = dict_gen if isinstance(tmp, dict) else list_gen
                for j, tmp_j in gen_2(tmp):
                    ret[fun(i, j)] = tmp_j
            else:
                ret[i] = tmp
        return ret
    return data


def data_index(data, key):
    """Indexing data for key or a list of keys."""

    def idx(data, i):
        if isinstance(i, int):
            return data[i]
        assert isinstance(data, dict)
        if i in data:
            return data[i]
        for k, v in data.items():
            if str(k) == str(i):
                return v
        raise ValueError("{} is not found".format(i))

    if isinstance(key, (list, tuple)):
        keys = list(key)
        if len(keys) > 1:
            return data_index(idx(data, keys[0]), keys[1:])
        return idx(data, keys[0])
    return idx(data, key)


def data_strip(data, keys):
    if isinstance(keys, str):
        keys = [keys]
    if isinstance(data, dict):
        ret = {}
        for k, v in data.items():
            if k not in keys:
                ret[k] = data_strip(v, keys)
        return ret
    if isinstance(data, list):
        return [data_strip(data_i, keys) for data_i in data]
    if isinstance(data, tuple):
        return tuple([data_strip(data_i, keys) for data_i in data])
    return data


def check_nan(data, no_raise=False):
    """check if there is nan in data"""
    head_keys = []

    def _check_nan(dat, head):
        if isinstance(dat, dict):
            return {k: _check_nan(v, head + [k]) for k, v in dat.items()}
        if isinstance(dat, list):
            return [
                _check_nan(data_i, head + [i]) for i, data_i in enumerate(dat)
            ]
        if isinstance(dat, tuple):
            return tuple(
                [
                    data_struct(data_i, head + [i])
                    for i, data_i in enumerate(dat)
                ]
            )
        if np.any(tf.math.is_nan(dat)):
            if no_raise:
                return False
            raise ValueError("nan in data[{}]".format(head))
        return True

    return _check_nan(data, head_keys)



#!/usr/bin/env python3
# from tf_pwa.phasespace import  PhaseSpaceGenerator
import os.path
import sys

import numpy as np

from tf_pwa.applications import gen_mc

this_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, this_dir + "/..")


"""
def flat_mc(number,outfile):
  a = PhaseSpaceGenerator(4.59925172,[2.01028,0.13957,2.00698]) #from data???
  flat_mc_data = a.generate(number)
  #print(w,flat_mc_data)
  #print(flat_mc_data[0].M())
  #print(flat_mc_data[1].M())
  #print(flat_mc_data[2].M())
  pb = flat_mc_data[0]
  pb_a = np.array([pb.T,pb.X,pb.Y,pb.Z]).reshape((4,-1))
  pc = flat_mc_data[1]
  pc_a = np.array([pc.T,pc.X,pc.Y,pc.Z]).reshape((4,-1))
  pd = flat_mc_data[2]
  pd_a = np.array([pd.T,pd.X,pd.Y,pd.Z]).reshape((4,-1))
  print(pd_a)

  pa = np.array([pd_a,pb_a,pc_a])
  print(pa.shape)
  pa = np.transpose(pa,(2,0,1)).reshape((-1,4))
  print(pa.shape)
  np.savetxt(outfile,pa)  # 一个不包含探测器效率的MC样本
"""
if __name__ == "__main__":
    data = gen_mc(4.59925172, [2.00698, 2.01028, 0.13957], 100, "tmp.dat")
    print(data)



"""
This module implements classes to describe particles and decays.
"""
import functools
from collections import UserList

import numpy as np

from .breit_wigner import barrier_factor as default_barrier_factor
from .cg import cg_coef
from .config import get_config
from .utils import deep_ordered_iter

# from pysnooper import snoop


DEFAULT_DECAY = "defalut_decay"


def cross_combine(x):
    """
    Combine every two of a list, as well as give every one of them.???
    Can be put to utils.py

    :param x:
    :return:
    """
    if not x:  # if x is []
        return []
    head = x[0]
    tail = x[1:]
    ret = []
    other = cross_combine(tail)
    for i in head:
        if not other:
            ret.append(i)
        else:
            for j in other:
                ret.append(i + j)
    return ret


def _spin_range(a, b):
    while a <= b:
        yield a
        a = a + 1


def _spin_int(x):
    if isinstance(x, int):
        return x
    return int(x + 0.1)


@functools.total_ordering
class BaseParticle(object):
    """
    Base Particle object. Name is "name[:id]".

    :param name: String. Name of the particle
    :param J: Integer or half-integer. The total spin
    :param P: 1 or -1. The parity
    :param spins: List. The spin quantum numbers. If it's not provided, ``spins`` will be ``tuple(range(-J, J + 1))``.
    :param mass: Real variable
    :param width: Real variable
    """

    def __init__(
        self,
        name,
        J=0,
        P=-1,
        C=None,
        spins=None,
        mass=None,
        width=None,
        id_=None,
        disable=False,
        **kwargs
    ):
        self.set_name(name, id_)
        self.decay = []  # list of Decay
        self.creators = []  # list of Decay which creates the particle
        if isinstance(J, str):
            J = eval(J)
        self.J = J
        self.P = P
        self.C = C
        if spins is None:
            spins = tuple(_spin_range(-J, J))
        self.spins = tuple(
            [eval(i) if isinstance(i, str) else i for i in spins]
        )
        self.mass = mass
        self.width = width
        self.disable = disable
        for k, v in kwargs.items():
            setattr(self, k, v)

    def set_name(self, name, id_=None):
        if id_ is None:
            names = name.split(":")
            if len(names) > 1:
                self._name = ":".join(names[:-1])
                try:
                    self._id = int(names[-1])
                except ValueError:
                    self._name, self._id = name, 0
            else:
                self._name, self._id = name, 0
        else:
            self._name, self._id = name, id_

    @property
    def name(self):
        return self._name

    def add_decay(self, d):
        """
        :param d: BaseDecay object
        """
        if d not in self.decay:
            self.decay.append(d)

    def remove_decay(self, d):
        """
        :param d: BaseDecay object
        """
        self.decay.remove(d)

    def add_creator(self, d):
        """
        Add a decay reaction where the particle is created.

        :param d: BaseDecay object
        """
        self.creators.append(d)

    def __repr__(self):
        if self._id == 0:
            return self._name
        return "{}:{}".format(self.name, self._id)

    def __hash__(self):
        return hash((self.name, self._id))

    def __eq__(self, other):
        if not isinstance(other, BaseParticle):
            return False
        return (self.name, self._id) == (other.name, other._id)

    def __lt__(self, other):
        if isinstance(other, BaseParticle):
            return (self.name, self._id) < (other.name, other._id)
        return self.__repr__() < other

    def __add__(self, other):
        return ParticleList([self]) + other

    def __lshift__(self, other):
        _BaseDecay, kwargs = get_config(DEFAULT_DECAY, (BaseDecay, {}))
        return _BaseDecay(self, other, **kwargs)

    def chain_decay(self):
        """return all decay chain self decay to"""
        ret = []
        for i in self.decay:
            ret_tmp = [[[i]]]
            for j in i.outs:
                tmp = j.chain_decay()
                if tmp:  # if tmp is not []
                    ret_tmp.append(tmp)
            ret += cross_combine(ret_tmp)
        return ret

    def get_resonances(self):
        """return all resonances self decay to"""
        decay_chain = self.chain_decay()
        chains = [DecayChain(i) for i in decay_chain]
        decaygroup = DecayGroup(chains)
        return decaygroup.resonances


class ParticleList(UserList):
    """List for Particle"""

    def __add__(self, other):
        if isinstance(other, BaseParticle):
            other = [other]
        return super(ParticleList, self).__add__(other)


def GetA2BC_LS_list(
    ja, jb, jc, pa=None, pb=None, pc=None, p_break=False, ca=None
):
    """
    The :math:`L-S` coupling for the decay :math:`A\\rightarrow BC`, where :math:`L` is the orbital
    angular momentum of :math:`B` and :math:`B`, and :math:`S` is the superposition of their spins.
    It's required that :math:`|J_B-J_C|\leq S \leq J_B+J_C` and :math:`|L-S|\leq J_A \leq L+S`. It's also required by the conservation of
    P parity that :math:`L` is keep :math:`P_A = P_B P_C  (-1)^{l}`.

    :param ja: `J` of particle `A`
    :param jb: `J` of particle `B`
    :param jc: `J` of particle `C`
    :param pa: `P` of particle `A`
    :param pb: `P` of particle `B`
    :param pc: `P` of particle `C`
    :param p_break: allow p voilate
    :param ca: enabel c partity select c=(-1)^(l+s)
    :return: List of :math:`(l,s)` pairs.
    """
    if pa is None or pb is None or pc is None:
        p_break = True
    if not p_break:
        dl = 0 if pa * pb * pc == 1 else 1  # pa = pb * pc * (-1)^l
    s_min = abs(jb - jc)
    s_max = jb + jc
    # ns = s_max - s_min + 1
    ret = []
    for s in _spin_range(s_min, s_max):
        for l in _spin_range(abs(ja - s), ja + s):
            if l % 1.0 < 0.7 and l % 1.0 > 0.3:  # half int
                break
            l = _spin_int(l)
            if ca is not None:
                if ca != (-1) ** (l + s):
                    continue
            if not p_break:
                if l % 2 == dl:
                    ret.append((l, s))
            else:
                ret.append((l, s))
    # print(ret, ja, jb, jc, pa, pb, pc)
    return ret


def simple_cache_fun(f):
    """

    :param f:
    :return:
    """
    name = "simple_cached_" + f.__name__

    @functools.wraps(f)
    def g(self):
        if not hasattr(self, name):
            setattr(self, name, f(self))
        return getattr(self, name)

    return g


@functools.total_ordering
class BaseDecay(object):
    """
    Base Decay object

    :param core: Particle. The mother particle
    :param outs: List of Particle. The daughter particles
    :param name: String. Name of the decay
    :param disable: Boolean. If it's True???
    """

    def __init__(
        self,
        core,
        outs,
        name=None,
        disable=False,
        p_break=False,
        c_break=True,
        curve_style=None,
        **kwargs
    ):
        self._name = name
        self.core = core
        self.outs = tuple(outs)
        self.p_break = p_break
        self.c_break = c_break
        if hasattr(self.core, "disable") and self.core.disable:
            disable = True
        for i in self.outs:
            if hasattr(i, "disable") and i.disable:
                disable = True
        if not disable:
            self.core.add_decay(self)
            for i in outs:
                i.add_creator(self)
        self.curve_style = curve_style
        for k, v in kwargs.items():
            setattr(self, k, v)

    @property
    def name(self):
        return self._name

    def __repr__(self):
        ret = str(self.core)
        ret += "->"
        ret += "+".join([str(i) for i in self.outs])
        return ret  # "A->B+C"

    @simple_cache_fun  # @functools.lru_cache()
    def get_id(self):
        return (self.core, tuple(sorted(self.outs)))

    def __hash__(self):
        return hash(self.get_id())

    def __eq__(self, other):
        if not isinstance(other, BaseDecay):
            return False
        return self.get_id() == other.get_id()

    def __lt__(self, other):
        if not isinstance(other, BaseDecay):
            return False
        return self.get_id() < other.get_id()


class Decay(BaseDecay):  # add useful methods to BaseDecay
    """
    General Decay object
    """

    @simple_cache_fun  # functools.lru_cache()
    def get_ls_list(self):
        """
        It has interface to ``tf_pwa.particle.GetA2BC_LS_list(ja, jb, jc, pa, pb, pc)``
        :return: List of (l,s) pairs
        """
        ca = self.core.C if not self.c_break else None
        ja = self.core.J
        jb = self.outs[0].J
        jc = self.outs[1].J
        pa = self.core.P
        pb = self.outs[0].P
        pc = self.outs[1].P
        return tuple(
            GetA2BC_LS_list(
                ja, jb, jc, pa, pb, pc, p_break=self.p_break, ca=ca
            )
        )

    # @functools.lru_cache()
    def get_l_list(self):
        """
        List of l in ``self.get_ls_list()``
        """
        return tuple([l for l, s in self.get_ls_list()])

    @functools.lru_cache()
    def get_min_l(self):
        """
        The minimal l in the LS coupling
        """
        return min(self.get_l_list())

    def generate_params(self, name=None, _ls=True):
        """
        Generate the name of the variable for every (l,s) pair. In PWA, the variable is usually expressed as
        :math:`g_ls`.

        :param name: String. It is the name of the decay by default
        :param _ls: ???
        :return: List of strings
        """
        if name is None:
            name = self.name
        ret = []
        for l, s in self.get_ls_list():
            name_r = "{name}_l{l}_s{s}_r".format(name=name, l=l, s=s)
            name_i = "{name}_l{l}_s{s}_i".format(name=name, l=l, s=s)
            ret.append((name_r, name_i))
        return ret

    @functools.lru_cache()
    def get_cg_matrix(self):  # CG factor inside H
        """
        The matrix indexed by :math:`[(l,s),(\\lambda_b,\\lambda_c)]`. The matrix element is

        .. math::
            \\sqrt{\\frac{ 2 l + 1 }{ 2 j_a + 1 }}
            \\langle j_b, j_c, \\lambda_b, - \\lambda_c | s, \\lambda_b - \\lambda_c \\rangle
            \\langle l, s, 0, \\lambda_b - \\lambda_c | j_a, \\lambda_b - \\lambda_c \\rangle

        This is actually the pre-factor of :math:`g_ls` in the amplitude formula.

        :return: 2-d array of real numbers
        """
        ls = self.get_ls_list()
        m = len(ls)
        ja = self.core.J
        jb = self.outs[0].J
        jc = self.outs[1].J
        n = (2 * jb + 1) * (2 * jc + 1)
        ret = np.zeros(shape=(n, m))
        for i, ls_i in enumerate(ls):
            l, s = ls_i
            j = 0
            for lambda_b in range(-jb, jb + 1):
                for lambda_c in range(-jc, jc + 1):
                    ret[j][i] = (
                        np.sqrt((2 * l + 1) / (2 * ja + 1))
                        * cg_coef(
                            jb, jc, lambda_b, -lambda_c, s, lambda_b - lambda_c
                        )
                        * cg_coef(
                            l,
                            s,
                            0,
                            lambda_b - lambda_c,
                            ja,
                            lambda_b - lambda_c,
                        )
                    )
                    j += 1
        return ret

    def barrier_factor(self, q, q0):  # Barrier factor inside H
        """
        The cached barrier factors with :math:`d=3.0` for all :math:`l`. For barrier factor, refer to
        **tf_pwa.breit_wigner.Bprime(L, q, q0, d)**

        :param q: Data array
        :param q0: Real number
        :return: 1-d array for every data point
        """
        d = 3.0
        ret = default_barrier_factor(self.get_l_list(), q, q0, d)
        return ret


def split_particle_type_list(decays):
    """
    Separate initial particle, intermediate particles, final particles in a decay chain.

    :param decays: DecayChain
    :return: Set of initial Particle, set of intermediate Particle, set of final Particle
    """
    core_particles = []
    out_particles = []

    for i in decays:
        core_particles.append(i.core)
        for j in i.outs:
            out_particles.append(j)

    inner = [
        i for i in core_particles if i in out_particles
    ]  # core_particles & out_particles
    top = [
        i for i in core_particles if i not in inner
    ]  # core_particles - inner
    outs = [
        i for i in out_particles if i not in inner
    ]  # out_particles - inner
    # print(decays, inner, top, outs)
    return top, inner, outs


def split_particle_type(decays):
    top, inner, outs = split_particle_type_list(decays)
    return set(top), set(inner), set(outs)


def split_len(dicts):
    """
    Split a dictionary of lists by their length.
    E.g. {"b":[2],"c":[1,3],"d":[1]} => [None, [('b', [2]), ('d', [1])], [('c', [1, 3])]]

    Put to utils.py or _split_len if not used anymore???

    :param dicts: Dictionary
    :return: List of dictionary
    """
    size_table = []
    for i in dicts:
        tmp = dicts[i]
        size_table.append((len(tmp), i))
    max_l = max([i for i, _ in size_table])
    ret = [None] * (max_l + 1)
    for i, s in size_table:
        if ret[i] is None:
            ret[i] = []
        ret[i].append((s, dicts[s]))
    return ret


class DecayChain(object):
    """
    A decay chain. E.g. :math:`A\\rightarrow BC, B\\rightarrow DE`

    :param chain: ???
    """

    def __init__(self, chain):
        self.chain = chain
        top, inner, outs = split_particle_type_list(chain)
        assert len(top) == 1, "top particles must be only one particle"
        self.top = top.pop()
        self.inner = sorted(list(inner))
        self.outs = sorted(list(outs))

    def __iter__(self):
        return iter(self.chain)

    def __getitem__(self, idx):
        if isinstance(idx, int):
            return self.chain[idx]
        raise NotImplementedError

    def __repr__(self):
        return "{}".format(list(self.chain))

    @simple_cache_fun  # functools.lru_cache()
    def sorted_table(self):
        """
        A topology independent structure.
        E.g. [a->rb,r->cd] => {a:[b,c,d],r:[c,d],b:[b],c:[c],d:[d]}

        :return: Dictionary indexed by Particle
        """
        decay_dict = {}
        for i in self.outs:
            decay_dict[i] = [i]

        chain = self.chain
        while chain:
            tmp_chain = []
            for i in chain:
                if all([j in decay_dict for j in i.outs]):
                    decay_dict[i.core] = []
                    for j in i.outs:
                        decay_dict[i.core] += decay_dict[j]
                    decay_dict[i.core].sort()
                else:
                    tmp_chain.append(i)
            chain = tmp_chain
        decay_dict[self.top] = sorted(list(self.outs))
        return decay_dict

    @simple_cache_fun
    def sorted_table_layers(self):
        """
        Get the layer of decay chain as sorted table.
        Or the list of particle with the same number of final particles.
        So, the first item is always None.
        E.g. [a->rb,r->cd] => [None, [(b, [b]), (c, [c]), (d, [d])], [(r, [c, d])], [(a, [b, c, d])]]

        :return: List of dictionary
        """
        st = self.sorted_table()
        return split_len(st)

    @staticmethod
    def from_sorted_table(decay_dict):
        """
        Create decay chain from a topology independent structure.
        E.g. {a:[b,c,d],r:[c,d],b:[b],c:[c],d:[d]} => [a->rb,r->cd]

        :param decay_dict: Dictionary
        :return: DecayChain
        """

        def sum_list(ls):
            ret = ls[0]
            for i in ls[1:]:
                ret = ret + i
            return ret

        def deep_search(idx, base):
            base_step = 2
            max_step = len(base)
            found = False
            while base_step <= max_step:
                for i in deep_ordered_iter(base, base_step):
                    check = sum_list([base[j] for j in i])
                    if sorted(check) == sorted(idx[1]):
                        found = True
                        return i
                base_step += 1
            if not found:
                raise Exception("not found in searching")

        s_dict = split_len(decay_dict)
        base_dict = dict(s_dict[1])
        ret = []
        for s_dict_i in s_dict[2:]:
            if s_dict_i:
                for j in s_dict_i:
                    found = deep_search(j, base_dict)
                    ret.append(BaseDecay(j[0], found, disable=True))
                    for i in found:
                        del base_dict[i]
                    base_dict[j[0]] = j[1]
        return DecayChain(ret)

    @staticmethod
    def from_particles(top, finals):
        """
        Build possible decay chain Topology.
        E.g. a -> [b,c,d] => [[a->rb,r->cd],[a->rc,r->bd],[a->rd,r->bc]]

        :param top: Particle
        :param finals: List of Particle
        :return: DecayChain
        """
        assert len(finals) > 0, " "

        def get_graphs(g, ps):
            if ps:
                p = ps[0]
                ps = ps[1:]
                ret = []
                for i in g.edges:
                    gi = g.copy()
                    gi.add_node(i, p)
                    ret += get_graphs(gi, ps)
                return ret
            return [g]

        base = _Chain_Graph()
        base.add_edge(top, finals[0])
        gs = get_graphs(base, finals[1:])
        return [
            gi.get_decay_chain(top, head="chain{}_".format(i))
            for i, gi in enumerate(gs)
        ]

    @functools.lru_cache()
    def topology_id(self, identical=True):
        """
        topology identy

        :param identical: allow identical particle in finals
        :return:
        """
        a = self.sorted_table()
        if identical:
            set_a = [[j.name for j in a[i]] for i in a]
        else:
            set_a = [list(a[i]) for i in a]
        return sorted(set_a)

    def __eq__(self, other):
        if not isinstance(other, DecayChain):
            return False
        return self.get_id() == other.get_id()

    def __lt__(self, other):
        if not isinstance(other, DecayChain):
            return False
        return self.get_id() < other.get_id()

    @simple_cache_fun
    def __hash__(self):
        return hash(self.get_id())

    @simple_cache_fun
    def get_id(self):
        """return identity of the decay"""
        decay = sorted(list(self.chain))
        return tuple(decay)

    def standard_topology(self):
        """
        standard topology structure of the decay chain,
        all inner particle will be replace as a tuple of out particles.
        for example [A->R+C, R->B+D], is [A->(B, D)+C, (B, D)->B+D]
        """
        a = self.sorted_table()
        name_map = {}
        for k, v in a.items():
            parts = sorted([str(i) for i in v])
            name_map[k] = "({})".format(", ".join(parts))
        name_map[self.top] = str(self.top)
        for i in self.outs:
            name_map[i] = str(i)
        particle_map = {k: BaseParticle(v) for k, v in name_map.items()}
        ret = []
        for i in self:
            core = particle_map[i.core]
            ret.append(BaseDecay(core, [particle_map[j] for j in i.outs]))
        return DecayChain(ret)

    def topology_map(self, other=None):
        """Mapping relations of the same topology decay
        E.g. [A->R+B,R->C+D],[A->Z+B,Z->C+D] => {A:A,B:B,C:C,D:D,R:Z,A->R+B:A->Z+B,R->C+D:Z->C+D}
        """
        a = self.sorted_table()
        if other is None:
            other = self.standard_topology()
        b = other.sorted_table()
        ret = {}
        for i in a:
            for j in b:
                if a[i] == b[j]:
                    ret[i] = j
                    break
        for i in self:
            test_decay = BaseDecay(
                ret[i.core], [ret[k] for k in i.outs], disable=False
            )
            for j in other:
                if test_decay == j:
                    ret[i] = j
                    break
        return ret

    def topology_same(self, other, identical=True):
        """
        whether self and other is the same topology

        :param other: other decay chains
        :param identical: using identical particles
        :return:
        """
        if not isinstance(other, DecayChain):
            raise TypeError("unsupport type {}".format(type(other)))
        return self.topology_id(identical) == other.topology_id(identical)

    def get_all_particles(self):
        """
        get all particles in the decay chains
        """
        ret = []
        for i in self:
            if i.core not in ret:
                ret.append(i.core)
            for j in i.outs:
                if j not in ret:
                    ret.append(j)
        return ret

    def get_particle_decay(self, particle):
        """
        get particle decay in the chain
        """
        p = BaseParticle(str(particle))
        for i in self:
            if i.core == p:
                return i
        raise ValueError("Not found particle {} decay".format(p))

    def depth_first(self, node_first=True):
        """
        depth first travel for decay
        """
        node_map = {i.core: i for i in self.chain}

        def _dep(t, d):
            dec = node_map.get(t, None)
            if dec is not None:
                if node_first:
                    yield d, dec
                for i in dec.outs:
                    yield from _dep(i, d + 1)
                if not node_first:
                    yield d, dec

        return _dep(self.top, 1)


class _Chain_Graph(object):
    def __init__(self):
        self.nodes = []
        self.edges = []
        self.count = 0

    def add_edge(self, a, b):
        self.edges.append((a, b))

    def add_node(self, e, d):
        self.edges.remove(e)
        count = self.count
        node = "node_{}".format(count)
        self.nodes.append(node)
        self.edges.append((e[0], node))
        self.edges.append((node, e[1]))
        self.edges.append((node, d))
        self.count += 1

    def copy(self):
        ret = _Chain_Graph()
        ret.nodes = self.nodes.copy()
        ret.edges = self.edges.copy()
        ret.count = self.count
        return ret

    def get_decay_chain(self, top, head="tmp_"):
        decay_list = {}
        ret = []
        inner_particle = {}
        for i in self.nodes:
            inner_particle[i] = BaseParticle("{}{}".format(head, i))
        for i, j in self.edges:
            i = inner_particle.get(i, i)
            j = inner_particle.get(j, j)
            if i in decay_list:
                decay_list[i].append(j)
            else:
                decay_list[i] = [j]
        assert len(decay_list[top]) == 1, ""
        tmp = decay_list[top][0]
        decay_list[top] = decay_list[tmp]
        del decay_list[tmp]
        for i in decay_list:
            tmp = BaseDecay(i, decay_list[i], disable=True)
            ret.append(tmp)
        return DecayChain(ret)


class DecayGroup(object):
    """
    A group of two-body decays.

    :param chains: List of DecayChain
    """

    def __init__(self, chains):
        first_chain = chains[0]
        if not isinstance(first_chain, DecayChain):
            chains = [DecayChain(i) for i in chains]
            first_chain = chains[0]
        self.chains = chains
        self.top = first_chain.top
        self.outs = sorted(list(first_chain.outs))
        for i in chains:
            assert i.top == first_chain.top, ""
            assert (
                i.outs == first_chain.outs
            ), "{} and {} particles are different".format(i, first_chain)
        # resonances = set()
        resonances = []
        for i in chains:
            # resonances |= i.inner
            for j in i.inner:
                if j not in resonances:
                    resonances.append(j)
        self.resonances = list(resonances)
        self.identical_particles = []

    def __repr__(self):
        return "{}".format(self.chains)

    def __iter__(self):
        return iter(self.chains)

    def __getitem__(self, idx):
        if isinstance(idx, int):
            return self.chains[idx]
        raise NotImplementedError

    def get_decay_chain(self, id_):
        """get decay chain from identity string"""
        if isinstance(id_, int):
            return self.chains[id_]
        if isinstance(id_, str):
            names = id_.split("/")
        elif isinstance(id_, list):
            names = [str(i) for i in id_]
        elif isinstance(id_, BaseParticle):
            names = [str(id_)]
        else:
            raise NotImplementedError
        return self.get_chain_from_particle(names)

    def topology_structure(self, identical=False, standard=True):
        """

        :param identical:
        :param standard:
        :return:
        """
        ret = []
        for i in self:
            for j in ret:
                if i.topology_same(j, identical):
                    break
            else:
                ret.append(i)
        if standard:
            return [i.standard_topology() for i in ret]
        return ret

    @functools.lru_cache()
    def get_chains_map(self, chains=None):
        """

        :param chains:
        :return:
        """
        if chains is None:
            chains = self.chains
        chain_maps = []
        for decays in self.topology_structure():
            decay_chain = DecayChain(list(decays))
            tmp = {}
            for j in chains:
                if decay_chain.topology_same(j):
                    chain_map = decay_chain.topology_map(j)
                    tmp[j] = chain_map
            chain_maps.append(tmp)
        return chain_maps

    def get_chain_from_particle(self, names):
        """
        get the first decay chain has all particles in names
        """
        particles = [BaseParticle(str(i)) for i in names]
        for decay_chain in self.chains:
            all_particles = list(decay_chain.get_all_particles())
            found = True
            for i in particles:
                if i not in all_particles:
                    found = False
                    break
            if found:
                return decay_chain
        raise ValueError("Not found such decay chain: {}".format(names))

    def get_particle(self, name):
        """
        get particle by name
        """
        for decay_chain in self.chains:
            all_particles = list(decay_chain.get_all_particles())
            for i in all_particles:
                if str(i) == str(name):
                    return i
        raise ValueError("Not found such decay chain: {}".format(name))
